#! usr/env/bin python3
''' Plot a normal model against a mesh

long description
'''

######## script ########

def plot_mesh_norm_corner(
                          fname_mesh,
                          fname_nal,
                          fname_plot,
                          label_mesh,
                          label_nal,
                          title=None,
                          sample_labels=None,
                          scale=2,
                          linewidth_scale=4.,
                          coordinate_labels=None,
                          **kwargs
                         ):
    ''' Plot a normal model against a mesh

    Parameters
    ----------
    fname_mesh: str
        Input file locations for mesh
    fname_nal: str
        Input file locations for nal
    fname_plot: str
        Input file location for plot
    label_mesh: str
        Input label for mesh
    label_nal: str
        Input label for nal
    title: str, optional
        Input title of plot
    sample_labels: list, optional
        Input legend labels for mesh and nal
    scale: float, optional
        Input width compared to PRD standard
    '''
    from gwalk.density.mesh import Mesh
    from gwalk.bounded_multivariate_normal import MultivariateNormal
    from gwalk.plots.likelihood_corner import corner_cross_sample_normal
    # load mesh
    mesh = Mesh.load(fname_mesh,label_mesh)
    # Load normal
    MV = MultivariateNormal.load(fname_nal,label_nal)
    if not coordinate_labels is None:
        for i in range(len(coordinate_labels)):
            MV._parameters["mu_%d"%i].label = coordinate_labels[i]
    # Generate the plot
    corner_cross_sample_normal(
                               fname_plot,
                               MV,
                               mesh,
                               title=title,
                               sample_labels=sample_labels,
                               scale=scale,
                               linewidth_scale=linewidth_scale,
                               **kwargs
                              )

def plot_GW(
            fname_mesh,
            fname_nal,
            fname_plot,
            event,
            coord_tag,
            group,
            fit_method,
            **kwargs
           ):
    '''Standardize some of the plot arguments
    Parameters
    ----------
    fname_mesh: str
        Input file locations for mesh
    fname_nal: str
        Input file locations for nal
    fname_plot: str
        Input file location for plot
    event: str
        Input GW event name
    coord_tag: str
        Input coordinate tag for fit
    group: str
        Input approximate for fit
    fit_method: str
        Input method of fit
    '''
    from gwalk.catalog.coordinates import coord_tags, coord_labels
    coordinate_labels = []
    for coord in coord_tags[coord_tag][:-1]:
        coordinate_labels.append(coord_labels[coord])
    plot_mesh_norm_corner(
                          fname_mesh,
                          fname_nal,
                          fname_plot,
                          "%s/%s:%s"%(event,coord_tag,group),
                          "%s/%s:%s:%s"%(event,coord_tag,group,fit_method),
                          title=event,
                          sample_labels=["Catalog Samples","NAL"],
                          coordinate_labels=coordinate_labels,
                          **kwargs
                         )

######### Main ########

######## Execution ########
if __name__ == "__main__":
    import sys
    from os.path import join
    if len(sys.argv) > 1:
        fname_mesh, fname_nal, fname_plot, event, coord_tag, group, fit_method = sys.argv[1:8]
    else:
        event = "GW190425"
        release = "GWTC-2"
        coord_tag = "full_precessing_tides"
        group = "IMRPhenomPv2_NRTidal-LS"
        #group = "PublicationSamples"
        extension="png"
        fit_method = "select"
        run = "/home/xevra/Event_Likelihood_Approximation/nal-runs/run_0002"
        fname_mesh = join(run, "likelihood_mesh.hdf5")
        fname_nal = join(run, "%s.nal.hdf5"%release)
        fname_plot = join(run,"figures",event,"%s:%s:%s:%s_likelihood.%s"%(
            event,coord_tag,group,fit_method,extension))
        scale=16.
        rot_xlabel=True
        tight=True
        legend_loc=[0.58,0.7]
        #legend_loc=[0.45,0.68]

    plot_GW(
            fname_mesh,
            fname_nal,
            fname_plot,
            event,
            coord_tag,
            group,
            fit_method,
            scale=scale,
            rot_xlabel=rot_xlabel,
            tight=tight,
            legend_loc=legend_loc
           )



#!/usr/env/bin python3

######## Setup ########
import numpy as np
import time

######## Globals ########
MAX_NDIM = 100
NDIM = 11
NGAUSS = int(1e3)
SEED = 42
RS = np.random.RandomState(SEED)

#NDIM = 3
#NSAMP = int(1e5)
#NGAUSS = int(1e2)
#LIMITS = np.tile([0.,1.],(NDIM,1))

#### Generate scale ####
#SCALE = np.ones((NGAUSS,NDIM))

#### Generate sample points ####
#SAMPLE_POINTS = np.empty((NSAMP, NDIM))
#for _i in range(NSAMP):
#    SAMPLE_POINTS[_i] = RS.uniform(size=NDIM)

#### Generate sample ####
#MU = np.empty((NGAUSS, NDIM))
#for _i in range(NGAUSS):
#    MU[_i] = RS.uniform(size=NDIM,low=LIMITS[:,0],high=LIMITS[:,1])

#### Generate COV ####
#COV = np.tile(np.diag(np.ones(NDIM)),(NGAUSS, 1, 1))

######## Functions ########


######## C extensions ########

#def cext_maha():
#    from gwalk.multivariate_normal import maha as gwalk_maha
#    ## Generate data ##
#    Y = SAMPLE_POINTS
#    X_mu = MU
#    scale = SCALE
#    U = U_of_cov(COV)
#    # Start timer 
#    t0 = time.time()
#    maha = gwalk_maha(X_mu, scale, U, Y)
#    # End timer
#    t1 = time.time()
#    print("  C Extension time:\t%f seconds!"%(t1-t0))
#    return maha

######## Numpy functions ########

#def numpy_maha():
#    ## Generate data ##
#    Y = SAMPLE_POINTS
#    X_mu = MU
#    scale = SCALE
#    U = U_of_cov(COV)
#    # Start timer 
#    t0 = time.time()
#    ## Calculate maha ##
#    maha = np.sum(
#        np.square(
#            np.sum(
#                   (Y[None,:,:,None]/scale[:,None,:,None] - X_mu[:,None,:,None])*
#                        U[:,None,:,:],
#                    axis=-2,
##                   )
#                 ),
#                 axis=-1,
#                )
#    # End timer
#    t1 = time.time()
#    print("  Numpy time:\t\t%f seconds!"%(t1-t0))
#    return maha

######## Tests ########

#def test_maha():
#    print("Testing mahalanobis distance:")
#    maha1 = numpy_maha()
#    maha2 = cext_maha()
#    assert np.allclose(maha1, maha2)
#    print("  cext pass!")

def test_nparams():
    ''' Test nparams_of_ndim and ndim_of_nparams
    '''
    from gwalk.multivariate_normal.decomposition import nparams_of_ndim
    from gwalk.multivariate_normal.decomposition import ndim_of_nparams
    print("Testing nparams_of_ndim and ndim_of_nparams.")
    for ndim in range(MAX_NDIM):
        assert ndim_of_nparams(nparams_of_ndim(ndim)) == ndim
    print("  pass!")

def test_random_gauss(ngauss=10, ndim=4):
    ''' Test random gaussian generation and repeatability
    '''
    from gwalk.multivariate_normal import random_gauss_params
    print("Testing random_gauss_params")
    RS.seed(SEED)
    X1 = random_gauss_params(ngauss, ndim, rs=RS)
    RS.seed(SEED)
    X2 = random_gauss_params(ngauss, ndim, rs=RS)
    assert np.allclose(X1, X2)
    print("  pass!")

def test_mu_std_of_params(ngauss=100, ndim=11):
    ''' Test mu_of_params
    '''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition import mu_of_params
    from gwalk.multivariate_normal.decomposition import std_of_params
    print("Testing mu_of_params and std_of_params.")
    RS.seed(SEED)
    X1 = random_gauss_params(ngauss, ndim, rs=RS)
    mu1 = mu_of_params(X1)
    std1 = std_of_params(X1)
    assert mu1.shape == (ngauss, ndim)
    assert std1.shape == (ngauss, ndim)
    print("  pass!")

def test_cor_of_params(ngauss=1000, ndim=11):
    ''' Test mu_of_params
    '''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition_numpy import cor_of_params as cor_numpy
    from gwalk.multivariate_normal.decomposition import cor_of_params as cor_cext

    #### Setup ####
    print("Testing cor_of_params.")
    RS.seed(SEED)
    X1 = random_gauss_params(ngauss, ndim, rs=RS)

    #### Numpy ####
    t0 = time.time()
    cor1 = cor_numpy(X1)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert cor1.shape == (ngauss, ndim, ndim)

    #### C Extension ####
    t0 = time.time()
    cor2 = cor_cext(X1)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert cor2.shape == (ngauss, ndim, ndim)
    assert np.allclose(cor1, cor2)
    print("  pass!")

def test_cov_of_params(ngauss=1000, ndim=11):
    ''' Test cov_of_params
    '''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition_numpy import cov_of_params as cov_numpy
    from gwalk.multivariate_normal.decomposition import cov_of_params as cov_cext

    #### Setup ####
    print("Testing cov_of_params.")
    RS.seed(SEED)
    X1 = random_gauss_params(ngauss, ndim, rs=RS)

    #### Numpy ####
    t0 = time.time()
    cov1 = cov_numpy(X1)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert cov1.shape == (ngauss, ndim, ndim)

    #### C Extension ####
    t0 = time.time()
    cov2 = cov_cext(X1)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert cov2.shape == (ngauss, ndim, ndim)
    assert np.allclose(cov1, cov2)
    print("  pass!")

def test_decompose(ngauss=1000, ndim=11):
    ''' Test mu_cov_of_params and offset_mu_cov_of_params'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition_numpy import mu_cov_of_params as mu_cov_numpy
    from gwalk.multivariate_normal.decomposition import mu_cov_of_params as mu_cov_cext
    from gwalk.multivariate_normal.decomposition_numpy import offset_mu_cov_of_params as decompose_numpy
    from gwalk.multivariate_normal.decomposition import offset_mu_cov_of_params as decompose_cext
    from gwalk.multivariate_normal.decomposition import params_of_offset_mu_std_cor

    #### Setup ####
    print("Testing mu_cov_of_params.")
    RS.seed(SEED)
    X1 = random_gauss_params(ngauss, ndim, rs=RS)
    nparams = X1.shape[1]

    #### Numpy ####
    t0 = time.time()
    mu1, cov1 = mu_cov_numpy(X1)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert mu1.shape == (ngauss, ndim)
    assert cov1.shape == (ngauss, ndim, ndim)

    #### C Extension ####
    t0 = time.time()
    mu2, cov2 = mu_cov_cext(X1)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert mu2.shape == (ngauss, ndim)
    assert cov2.shape == (ngauss, ndim, ndim)
    assert np.allclose(mu1, mu2)
    assert np.allclose(cov1, cov2)
    print("  pass!")

    #### Decompose ####
    print("Testing offset_mu_cov_of_params.")
    #### Numpy ####
    t0 = time.time()
    off1, mu1, cov1 = decompose_numpy(X1)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert off1.shape == (ngauss,)
    assert mu1.shape == (ngauss, ndim)
    assert cov1.shape == (ngauss, ndim, ndim)

    #### C Extension ####
    t0 = time.time()
    off2, mu2, cov2 = decompose_cext(X1)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert off2.shape == (ngauss,)
    assert mu2.shape == (ngauss, ndim)
    assert cov2.shape == (ngauss, ndim, ndim)
    assert np.allclose(off1, off2)
    assert np.allclose(mu1, mu2)
    assert np.allclose(cov1, cov2)
    print("  pass!")

def test_cov_std_cor(ngauss=1000, ndim=11):
    ''' Test cov_of_params
    '''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition import std_of_params, cor_of_params, cov_of_params
    from gwalk.multivariate_normal.decomposition_numpy import cov_of_std_cor as cov_of_std_cor_numpy
    from gwalk.multivariate_normal.decomposition_numpy import std_of_cov as std_of_cov_numpy
    from gwalk.multivariate_normal.decomposition_numpy import cor_of_cov as cor_of_cov_numpy
    from gwalk.multivariate_normal.decomposition import cov_of_std_cor as cov_of_std_cor_cext
    from gwalk.multivariate_normal.decomposition import std_of_cov as std_of_cov_cext
    from gwalk.multivariate_normal.decomposition import cor_of_cov as cor_of_cov_cext
    from gwalk.multivariate_normal.decomposition import cor_of_std_cov as cor_of_std_cov_cext

    #### Setup ####
    RS.seed(SEED)
    X = random_gauss_params(ngauss, ndim, rs=RS)
    std0 = std_of_params(X)
    cor0 = cor_of_params(X)
    cov0 = cov_of_params(X)

    print("Testing cov_of_std_cor.")

    #### Numpy ####
    t0 = time.time()
    cov1 = cov_of_std_cor_numpy(std0, cor0)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert cov1.shape == (ngauss, ndim, ndim)

    #### C Extension ####
    t0 = time.time()
    cov2 = cov_of_std_cor_cext(std0, cor0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert cov2.shape == (ngauss, ndim, ndim)
    assert np.allclose(cov1, cov2)

    print("Testing std_of_cov")

    #### Numpy ####
    t0 = time.time()
    std1 = std_of_cov_numpy(cov0)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert std1.shape == (ngauss, ndim)
    assert np.allclose(std1, std0)

    #### C Extension ####
    t0 = time.time()
    std2 = std_of_cov_cext(cov0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert std2.shape == (ngauss, ndim)
    assert np.allclose(std2, std0)

    print("Testing cor_of_cov")

    #### Numpy ####
    t0 = time.time()
    cor1 = cor_of_cov_numpy(cov0)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert cor1.shape == (ngauss, ndim, ndim)
    assert np.allclose(cor1, cor0)

    #### C Extension ####
    t0 = time.time()
    cor2 = cor_of_cov_cext(cov0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert cor2.shape == (ngauss, ndim, ndim)
    assert np.allclose(cor2, cor0)
    print("  pass!")

    print("Testing cor_of_std_cov")

    #### C Extension ####
    t0 = time.time()
    cor2 = cor_of_std_cov_cext(std0, cov0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert cor2.shape == (ngauss, ndim, ndim)
    assert np.allclose(cor2, cor0)
    print("  pass!")

def test_recompose(ngauss=1000, ndim=11):
    ''' Test mu_cov_of_params and offset_mu_cov_of_params'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition import mu_cov_of_params
    from gwalk.multivariate_normal.decomposition import offset_mu_cov_of_params
    from gwalk.multivariate_normal.decomposition import offset_of_params
    from gwalk.multivariate_normal.decomposition import mu_of_params
    from gwalk.multivariate_normal.decomposition import std_of_params
    from gwalk.multivariate_normal.decomposition import cor_of_params
    from gwalk.multivariate_normal.decomposition import cov_of_params
    from gwalk.multivariate_normal.decomposition import params_of_offset_mu_std_cor
    from gwalk.multivariate_normal.decomposition import params_of_offset_mu_cov as params_of_offset_mu_cov_cext
    from gwalk.multivariate_normal.decomposition_numpy import params_of_offset_mu_cov as params_of_offset_mu_cov_numpy
    from gwalk.multivariate_normal.decomposition import params_of_mu_cov as params_of_mu_cov_cext
    from gwalk.multivariate_normal.decomposition_numpy import params_of_mu_cov as params_of_mu_cov_numpy

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(ngauss, ndim, rs=RS)
    off0 = offset_of_params(X0)
    mu0 = mu_of_params(X0)
    std0 = std_of_params(X0)
    cor0 = cor_of_params(X0)
    cov0 = cov_of_params(X0)
    nparams = X0.shape[1]

    #### Recompose ####
    print("Testing params_of_offset_mu_std_cor")
    t0 = time.time()
    X1 = params_of_offset_mu_std_cor(off0, mu0, std0, cor0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert X1.shape == (ngauss, nparams)
    assert np.allclose(X1, X0)
    print("  pass!")

    print("Testing params_of_offset_mu_cov")

    #### Numpy ####
    t0 = time.time()
    X2 = params_of_offset_mu_cov_numpy(off0, mu0, cov0)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert X2.shape == (ngauss, nparams)
    assert np.allclose(X2, X0)

    #### C Extension ####
    t0 = time.time()
    X1 = params_of_offset_mu_cov_cext(off0, mu0, cov0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert X1.shape == (ngauss, nparams)
    assert np.allclose(X1, X0)
    print("  pass!")

    print("Testing params_of_mu_cov")

    #### Numpy ####
    t0 = time.time()
    X3 = params_of_mu_cov_numpy(mu0, cov0)
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    assert X3.shape == (ngauss, nparams)
    assert np.allclose(X3[:,1:], X0[:,1:])
    assert np.allclose(X3[:,0], 0.)

    #### C Extension ####
    t0 = time.time()
    X4 = params_of_mu_cov_cext(mu0, cov0)
    t1 = time.time()
    print("  C Extension time:\t%f"%(t1-t0))
    assert X4.shape == (ngauss, nparams)
    assert np.allclose(X4[:,1:], X0[:,1:])
    assert np.allclose(X4[:,0], 0.)
    print("  pass!")

def test_reduce(ngauss = 1000, ndim = 11):
    '''test reduce_params'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition import params_reduce

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(ngauss, ndim, rs=RS)

    #### Test time evaluation ####
    print("Testing params_reduce_dd")
    t0 = time.time()
    X1 = params_reduce(X0, [0,1])
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))

    #### Test time evaluation ####
    print("Testing params reduce 1-D")
    t0 = time.time()
    X2 = params_reduce(X0, [0])
    t1 = time.time()
    print("  Numpy time:\t\t%f"%(t1-t0))
    



######## Main ########
def main():
    ngauss = NGAUSS
    ndim = NDIM
    test_nparams()
    test_random_gauss(ngauss, ndim)
    test_mu_std_of_params(ngauss, ndim)
    test_cor_of_params(ngauss, ndim)
    test_cov_of_params(ngauss, ndim)
    test_decompose(ngauss, ndim)
    test_cov_std_cor(ngauss, ndim)
    test_recompose(ngauss, ndim)
    test_reduce(ngauss, ndim)
    return

######## Execution ########
if __name__ == "__main__":
    main()



#! Reorganize nal data products


######## Globals ########

RELEASE = "GWTC-3"
NALDATA = "/home/xevra/Repos/nal-data/"
OUTDATA = "/home/xevra/Projects/basil-tools/pipeline/NAL/"
APPROXIMANTS = [
                "PublicationSamples",
                "NRSur7dq4",
                "SEOBNRv4PHM",
                "SEOBNRv4P",
                "SEOBNRv3",
                "IMRPhenomXPHM",
                "IMRPhenomPv3HM",
                "IMRPhenomPv2",
                "IMRPhenomPv2NRT_highSpin",
                "IMRPhenomPv2NRT_lowSpin",
                "IMRPhenomPv2_NRTidal-HS",
                "IMRPhenomPv2_NRTidal-LS",
                "IMRPhenomNSBH",
                "IMRPhenomNSBH_HighSpin",
                "IMRPhenomNSBH_LowSpin",
               ]
COMPRESSION = "gzip"

COORD_TAGS = [
              "aligned3d_source",
              "aligned3d_dist",
             ]

FIT_METHODS = [
               "select",
              ]


######## Imports ########

import numpy as np
from gwalk.data import Database
from os.path import join, isfile

######## Initialization ########

fname_release = join(NALDATA, "%s.nal.hdf5"%RELEASE)
db_release = Database(fname_release)
fname_final = join(OUTDATA, "%s.nal.hdf5"%RELEASE)
db_final = Database(fname_final)

######## Initialize events and copy attrs ########

# Identify top level attrs
release_attrs = db_release.attr_dict(".")
# Copy top level attrs
db_final.attr_set_dict('.',release_attrs)

# Identify events
events = db_release.list_items()
# Begin event loop
for item in events:
    # Create event group
    db_final.create_group(item)
    # Load event attrs
    event_attrs = db_release.attr_dict(item)
    # Set event attrs
    db_final.attr_set_dict(item,event_attrs)
    # get labels
    all_labels = db_release.list_items(item)
    # Begin label loop
    for jtem in all_labels:
        # split label
        coord_tag, approximant, fit_method = jtem.split(":")
        # check coord tag
        if not (coord_tag in COORD_TAGS):
            continue
        # Check approximant
        if not (approximant in APPROXIMANTS):
            continue
        # Check fit method
        if not (fit_method in FIT_METHODS):
            continue
        # We have decided to keep this label
        # Initialize the group
        db_final.create_group(join(item,jtem))
        # check group attrs
        group_attrs = db_release.attr_dict(join(item,jtem))
        # Save group attrs
        db_final.attr_set_dict(join(item,jtem),group_attrs)
        # Check datasets
        dsets = db_release.list_items(join(item,jtem),kind='dset')
        # Loop the dsets
        for ktem in dsets:
            # copy each dset
            db_final.dset_set(
                              join(item,jtem,ktem),
                              db_release.dset_value(join(item,jtem,ktem)),
                              compression=COMPRESSION,
                             )
# Done!
print("Success!")



#!/usr/bin/env python3
'''\
Populate nal-figures with run figures and links
'''
import numpy
import os
from os.path import join, isfile, isdir

######## Globals ########
TOPDIR = "/home/xevra/Event_Likelihood_Approximation"
RUNDIR = join(TOPDIR, "nal-runs")
FIGDIR = join(TOPDIR, "nal-figures")

######## Copy plots ########
# Identify runs
runs = os.listdir(RUNDIR)
# Loop
for run in runs:
    # Make sure directory exists
    if not isdir(join(FIGDIR, run)):
        os.mkdir(join(FIGDIR, run))
    # Make sure images exists
    imagedir = join(FIGDIR, run, "images")
    if not isdir(imagedir):
        os.mkdir(imagedir)
    # rsync plots
    cmd = "rsync -ravz %s/* %s"%(
                                join(RUNDIR, run, "figures"),
                                imagedir
                               )
    os.system(cmd)

######## Gather information ########
# Move to the working directory 
os.chdir(FIGDIR)
# Initialize dictionaries for storing information about plots
allplots = {}
allevents = {}
allcoords = {}
allwaveforms = {}
allmethods = {}
# Loop through runs
for run in runs:
    # Initialize lists
    runplots = []
    runevents = []
    runcoords = []
    runwaveforms = []
    runmethods = []
    # Loop through plots
    for event in os.listdir(join(FIGDIR, run, "images")):
        for plot in os.listdir(join(FIGDIR, run, "images", event)):
            # Gather fname
            fname = join(RUNDIR, run, "figures", event, plot)
            runplots.append(fname)
            fields = plot.rstrip(".pdfng").rstrip("likelihood").rstrip("_").split(':')
            # Gather additional information
            runevents.append(fields[0])
            runcoords.append(fields[1])
            runwaveforms.append(fields[2])
            runmethods.append(fields[3])

    allplots[run] = runplots
    allevents[run] = runevents
    allcoords[run] = runcoords
    allwaveforms[run] = runwaveforms
    allmethods[run] = runmethods

######## Generate link structure using recursive symbolic links ########

for run in runs:
    #### Initialize directories ####
    ## sort1:
    ## coord
    ##  |_ waveform_method
    ## 
    ## sort2:
    ## coord
    ## |_ event

    # Name variables
    sort1 = "by_coord_and_waveform_method"
    sort2 = "by_coord_and_event"
    sortdir1 = join(FIGDIR, run, sort1)
    sortdir2 = join(FIGDIR, run, sort2)
    # Ensure directory exists
    if not isdir(sortdir1):
        os.mkdir(sortdir1)
    if not isdir(sortdir2):
        os.mkdir(sortdir2)
    # Loop each plot
    for i in range(len(allplots[run])):
        # Extract fields
        item = allplots[run][i].split("/")[-1]
        coord = allcoords[run][i]
        event = allevents[run][i]
        waveform = allwaveforms[run][i]
        method = allmethods[run][i]
        # check if coord directories exist
        if not isdir(join(sortdir1, coord)):
            os.mkdir(join(sortdir1, coord))
        if not isdir(join(sortdir2, coord)):
            os.mkdir(join(sortdir2, coord))
        # Ensure waveform_method directories exist
        plotdir1 = join(sortdir1, coord, "%s_%s"%(waveform, method))
        if not isdir(plotdir1):
            os.mkdir(plotdir1)
        plotdir2 = join(sortdir2, coord, event)
        if not isdir(plotdir2):
            os.mkdir(plotdir2)

        # Create a relative symbolic link to the appropriate plot
        # Sort1
        link_loc = join(run,sort1,coord,"%s_%s"%(waveform, method),item)
        if not isfile(link_loc):
            cmd = "ln -frs %s %s"%(
                                   join(run,"images",event,item),
                                   link_loc,
                                  )
            print(cmd)
            os.system(cmd)
        # Sort2
        link_loc = join(run,sort2,coord,event,item)
        if not isfile(link_loc):
            cmd = "ln -frs %s %s"%(
                                   join(run,"images",event,item),
                                   link_loc
                                  )
            print(cmd)
            os.system(cmd)




        
    




#!/usr/env/bin python3

######## Setup ########
import numpy as np
import time

######## Globals ########
MAX_NDIM = 100
NDIM = 3
NPTS = int(1e4)
NGAUSS = int(1e2)
SEED = 42
RS = np.random.RandomState(SEED)

def test_unscaled(ngauss, ndim, npts):
    ''' Test the pdf evaluation with scale == 1'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition import mu_cov_of_params
    from scipy.stats import multivariate_normal as scipy_norm
    #from gwalk.utils.multivariate_normal import multivariate_normal_pdf as gwalk_pdf_old
    from gwalk.multivariate_normal import pdf as gwalk_pdf

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(ngauss, ndim, rs=RS)
    X0[:,ndim + 1:2*ndim + 1] += np.sqrt(0.1)
    X0[:,2*ndim + 1:] = 0.
    mu, cov = mu_cov_of_params(X0)
    sample = RS.uniform(size=npts*ndim).reshape(npts, ndim)

    #### Evaluate the pdf ####
    print("Testing pdf evaluation")
    t0 = time.time()
    L1 = np.empty((ngauss, npts))
    for i in range(ngauss):
        L1[i] = scipy_norm.pdf(sample, mean=mu[i], cov=cov[i])#, allow_singular=True)
    t1 = time.time()
    lnL1 = np.log(L1)
    keep = np.isfinite(lnL1)
    print("  Scipy time:\t\t%f"%(t1-t0))

    #t0 = time.time()
    #L2 = gwalk_pdf_old(X0, sample, scale=np.ones(ndim), log_scale=False)
    #t1 = time.time()
    #print("  old gwalk time:\t%f"%(t1-t0))
    #assert np.allclose(L1, L2)
    #print("  old gwalk pass!")

    t0 = time.time()
    L3 = gwalk_pdf(mu, cov, sample, scale=None, log_scale=False)
    t1 = time.time()
    print("  gwalk time:\t\t%f"%(t1-t0))
    assert np.allclose(L1, L3)
    print("  gwalk pass!")

    #t0 = time.time()
    #lnL2 = gwalk_pdf_old(X0, sample, scale=np.ones(ndim), log_scale=True)
    #t1 = time.time()
    #print("  old gwalk log time:\t%f"%(t1-t0))
    #assert np.allclose(lnL1[keep], lnL2[keep])
    #print("  old gwalk log pass!")

    t0 = time.time()
    lnL3 = gwalk_pdf(mu, cov, sample, scale=None, log_scale=True)
    t1 = time.time()
    print("  gwalk log time:\t%f"%(t1-t0))
    assert np.allclose(lnL1[keep], lnL3[keep])
    print("  gwalk log pass!")

def test_scaled(ngauss, ndim, npts):
    ''' Test the pdf evaluation with scale == 1'''
    from gwalk.multivariate_normal import random_gauss_params
    from gwalk.multivariate_normal.decomposition import mu_cov_of_params
    from scipy.stats import multivariate_normal as scipy_norm
    #from gwalk.utils.multivariate_normal import multivariate_normal_pdf as gwalk_pdf_old
    from gwalk.multivariate_normal import pdf as gwalk_pdf
    from gwalk.multivariate_normal.decomposition import params_rescale

    #### Setup ####
    RS.seed(SEED)
    X0 = random_gauss_params(ngauss, ndim, rs=RS)
    X0[:,ndim + 1:2*ndim + 1] += np.sqrt(0.1)
    X0[:,2*ndim + 1:] = 0.
    scale = np.exp(2 * RS.uniform(size=ndim*ngauss) - 1).reshape(ngauss,ndim)
    mu0, cov0 = mu_cov_of_params(X0)
    sample = RS.uniform(size=npts*ndim).reshape(npts, ndim)
    X1 = params_rescale(X0, scale)
    mu1, cov1 = mu_cov_of_params(X1)

    #### Get the non-scaled version for comparision ####
    L0 = np.empty((ngauss, npts))
    for i in range(ngauss):
        L0[i] = scipy_norm.pdf(sample, mean=mu0[i], cov=cov0[i])#, allow_singular=True)
    lnL0 = np.log(L0)
    keep = np.isfinite(lnL0)

    #### Evaluate the pdf ####
    print("Testing scaled pdf evaluation")
    t0 = time.time()
    L1 = np.empty((ngauss, npts))
    for i in range(ngauss):
        L1[i] = np.prod(scale[i])*scipy_norm.pdf(sample*scale[i], mean=mu1[i], cov=cov1[i])#, allow_singular=True)
    t1 = time.time()
    print("  Scipy time:\t\t%f"%(t1-t0))
    assert np.allclose(L0,L1) 

    ## new gwalk ##
    t0 = time.time()
    L2 = gwalk_pdf(mu0, cov0, sample, scale=scale, log_scale=False)
    t1 = time.time()
    print("  gwalk time:\t\t%f"%(t1-t0))
    assert np.allclose(L0, L2)
    print("  pass!")

    #### Evaluate the log pdf ####
    print("Testing scaled log pdf evaluation")
    t0 = time.time()
    L1 = np.empty((ngauss, npts))
    for i in range(ngauss):
        L1[i] = np.prod(scale[i])*scipy_norm.pdf(sample*scale[i], mean=mu1[i], cov=cov1[i])#, allow_singular=True)
    lnL1 = np.log(L1)
    t1 = time.time()
    assert np.allclose(lnL1[keep],lnL0[keep]) 
    print("  Scipy time:\t\t%f"%(t1-t0))

    ## new gwalk ##
    t0 = time.time()
    lnL2 = gwalk_pdf(mu0, cov0, sample, scale=scale, log_scale=True)
    t1 = time.time()
    print("  gwalk time:\t\t%f"%(t1-t0))
    assert np.allclose(lnL2[keep],lnL0[keep]) 
    print("  pass!")

######## Main ########
def main():
    ngauss = NGAUSS
    ndim = NDIM
    npts = NPTS
    test_unscaled(ngauss, ndim, npts)
    test_scaled(ngauss, ndim, npts)
    return

######## Execution ########
if __name__ == "__main__":
    main()



''' Functions for optimizing on a grid'''
######## Imports ########
from gwalk.multivariate_normal import MultivariateNormal
from gwalk.multivariate_normal import offset_of_params
from gwalk.multivariate_normal.utils import eigvals_satisfy_constraints
from gwalk.multivariate_normal.utils import cov_eigvals_by_eps

######## Optimization ########

def kl_of_grid(
               X,
               MV,
               Y,
               P,
               lnP,
              ):
    ''' Calculate the KL Divergence for a set of parameters

    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input test params for kl divergence
    MV: MultivariateNormal object
        Input bounded multivariate normal object
    Y: array like, shape = (npts, ndim)
        These are your likelihood grid points
    P: array like, shape = (npts,)
        The grid likelihoods (linear scale)
    lnP: array like, shape = (npts,)
        The log of the grid likelihoods
    '''
    # Imports 
    import numpy as np
    from basil_core.stats.distance import rel_entr
    ## Calculate kl divergences ##
    # Get lnL_norm
    lnQ = MV.likelihood(
                        Y,
                        X=X,
                        log_scale=True,
                        dynamic_rescale=True,
                       )
    # Calculate the kl divergence
    # Normalization of Q is handled by relative entropy alt
    kl = rel_entr(P, lnP=lnP, lnQ=lnQ, normQ=True)
    return kl

def lstsq_of_grid(
                  X,
                  MV,
                  Y,
                  lnP,
                  scalar=True,
                 ):
    ''' Calculate the least squares difference for a set of parameters

    Parameters
    ----------
    X: array like, shape = (npts, nparams)
        Input test params for kl divergence
    MV: MultivariateNormal object
        Input bounded multivariate normal object
    Y: array like, shape = (npts, ndim)
        These are your likelihood grid points
    lnP: array like, shape = (npts,)
        The log of the grid likelihoods
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts, nparams),optional
            Input test params for kl divergence
        scalar: bool (otpional)
            Input decide if you want a single number output instead of an array
        '''
    # Imports 
    import numpy as np
    from basil_core.stats.distance import rel_entr
    # Get lnL_norm
    #lnQ = np.zeros((X.shape[0],self.npts))
    lnQ = MV.likelihood(
                        Y,
                        X=X,
                        log_scale=True,
                        dynamic_rescale=True,
                       )
    # lstsq
    lstsq = np.sum((lnP[None,:] - lnQ)**2,axis=1)
    if scalar:
        lstsq = np.prod(lstsq)
    return lstsq

######## Algorithm ########

def optimize_likelihood_grid(
                             Y,
                             lnP = None,
                             P = None,
                             objective="kl",
                             method="COBYLA",
                             **kwargs
                            ):
    '''Optimize a truncated gaussian to a grid of evaluations

    Parameters
    ----------
    Y: array like, shape = (npts, ndim)
        These are your likelihood grid points
    P: array like, shape = (npts,)
        The grid likelihoods (linear scale)
    lnP: array like, shape = (npts,)
        The log of the grid likelihoods
    objective: string
        Option for choice of objective function
    method: string
        Scipy optimizer method

    Returns
    -------
    MV: gwalk.MultivariateNormal object
        An optimized MV object with optimized parameters MV.params
    '''
    # Imports
    import numpy as np
    import scipy.optimize
    from gwalk.utils.multivariate_normal import mu_of_params
    from gwalk.utils.multivariate_normal import cov_of_params

    #### Handle inputs ####
    ## Handle simple inputs ##
    assert isinstance(Y, np.ndarray)
    assert len(Y.shape) == 2
    npts, ndim = Y.shape
    
    ## Handle P ##
    assert ((not (P is None)) or (not (lnP is None)))
    if isinstance(P, np.ndarray):
        assert P.size == npts
    if isinstance(lnP, np.ndarray):
        assert lnP.size == npts
    if P is None:
        P = np.exp(lnP)
    if lnP is None:
        lnP = np.log(P)

    ## Remove bad points ##
    keep = np.isfinite(lnP) & np.isfinite(P)
    Y =     Y[keep]
    P =     P[keep]
    lnP =   lnP[keep]

    ## Generate from samples ##
    MV = MultivariateNormal.from_samples(Y, **kwargs)

    ## Define the KL objective ##
    def kl_objective(X):
        kl = kl_of_grid(X, MV, Y, P, lnP)
        return kl
    def lstsq_objective(X):
        lstsq = lstsq_of_grid(X, MV, Y, lnP)
        return lstsq
        

    if objective == "kl":
        fn = kl_objective
    elif objective == "lstsq":
        fn = lstsq_objective
    else:
        raise RuntimeError("objective options: 'kl', 'lstsq'")

    ## Generate constraints ##
    def satisfies_constraints(params, i):
        # Get the ratio of the eigenvalues to the eps
        s = cov_eigvals_by_eps(params)
        return s[:,i]
    cons = []
    for i in range(ndim):
        cons.append({'type':'ineq', 'fun':
            lambda params : satisfies_constraints(params, i) - 1})
    cons = tuple(cons)

    # Do the optimization
    scipy_out = scipy.optimize.minimize(
                                        fn,
                                        MV.params,
                                        method=method,
                                        bounds=MV.plimits,
                                        constraints=cons,
                                       )

    # Extract parameters
    X_scipy = scipy_out.x.flatten()

    print(X_scipy)
    # Still need to handle normalization constant for kl divergence
    if objective == "kl":
        # Still need a least squares fit for the normalization constant
        def lstsq_norm(X0):
            X_norm = np.copy(X_scipy)
            X_norm[0] = X0
            X_norm = X_norm.reshape((1, X_norm.size))
            lstsq = lstsq_of_grid(X_norm, MV, Y, lnP)
            return lstsq
        # Minimize this one
        norm_scipy_out = scipy.optimize.minimize(
                                                 lstsq_norm,
                                                 offset_of_params(X_scipy),
                                                 method=method,
                                                 bounds=[MV.plimits[0]],
                                                 #constraints=cons,
                                                )
        X_scipy[0] = norm_scipy_out.x.flatten()
            
    print(eigvals_satisfy_constraints(X_scipy))
    print(cov_eigvals_by_eps(X_scipy))
    MV.params = X_scipy
    return MV



#!/usr/env/bin python3

######## Setup ########
import numpy as np
from scipy import stats
import time

######## Globals ########

NDIM = 3
NSAMP = int(1e3)
NGAUSS = int(1e2)
LIMITS = np.tile([0.,1.],(NDIM,1))
SEED = 42
RS = np.random.RandomState(SEED)

#### Generate scale ####
SCALE = np.ones((NGAUSS,NDIM))

#### Generate sample points ####
SAMPLE_POINTS = np.empty((NSAMP, NDIM))
for _i in range(NSAMP):
    SAMPLE_POINTS[_i] = RS.uniform(size=NDIM)

#### Generate sample ####
MU = np.empty((NGAUSS, NDIM))
for _i in range(NGAUSS):
    MU[_i] = RS.uniform(size=NDIM,low=LIMITS[:,0],high=LIMITS[:,1])

#### Generate COV ####
COV = np.tile(np.diag(np.ones(NDIM)),(NGAUSS, 1, 1))

######## Functions ########
def U_of_cov(cov):
    s, u = np.linalg.eigh(cov)
    eps = stats._multivariate._eigvalsh_to_eps(s)
    keep = np.prod(s > eps, axis=1).astype(bool)
    d = np.zeros_like(s)
    d[keep] = np.log(s[keep])
    log_det_cov = np.sum(d, axis=1)
    s_pinv = np.zeros_like(s)
    s_pinv[keep] = 1./np.sqrt(s[keep])
    U = u*s_pinv[:,None,:]
    return U

def log_det_cov_of_cov(cov):
    s, u = np.linalg.eigh(cov)
    eps = stats._multivariate._eigvalsh_to_eps(s)
    keep = np.prod(s > eps, axis=1).astype(bool)
    d = np.zeros_like(s)
    d[keep] = np.log(s[keep])
    log_det_cov = np.sum(d, axis=1)
    return log_det_cov


######## C extensions ########

def cext_maha():
    from gwalk.multivariate_normal import maha as gwalk_maha
    ## Generate data ##
    Y = SAMPLE_POINTS
    X_mu = MU
    scale = SCALE
    U = U_of_cov(COV)
    # Start timer 
    t0 = time.time()
    maha = gwalk_maha(X_mu, scale, U, Y)
    # End timer
    t1 = time.time()
    print("  C Extension time:\t%f seconds!"%(t1-t0))
    return maha

def cext_pdf_exp_product():
    from gwalk.multivariate_normal import pdf_exp_product as gwalk_pdf_exp_product
    from gwalk.multivariate_normal import maha as gwalk_maha
    Y = SAMPLE_POINTS
    X_mu = MU
    scale = SCALE
    U = U_of_cov(COV)
    maha = gwalk_maha(X_mu, scale, U, Y)
    log_det_cov = log_det_cov_of_cov(COV)
    # Start timer 
    t0 = time.time()
    L = gwalk_pdf_exp_product(maha, log_det_cov, NDIM)
    # End timer
    t1 = time.time()
    print("  C Extension time:\t%f seconds!"%(t1-t0))
    return L

######## Numpy functions ########

def numpy_maha():
    ## Generate data ##
    Y = SAMPLE_POINTS
    X_mu = MU
    scale = SCALE
    U = U_of_cov(COV)
    # Start timer 
    t0 = time.time()
    ## Calculate maha ##
    maha = np.sum(
        np.square(
            np.sum(
                   (Y[None,:,:,None]/scale[:,None,:,None] - X_mu[:,None,:,None])*
                        U[:,None,:,:],
                    axis=-2,
                   )
                 ),
                 axis=-1,
                )
    # End timer
    t1 = time.time()
    print("  Numpy time:\t\t%f seconds!"%(t1-t0))
    return maha

def numpy_pdf_exp_product():
    from gwalk.multivariate_normal import maha as gwalk_maha
    Y = SAMPLE_POINTS
    X_mu = MU
    scale = SCALE
    U = U_of_cov(COV)
    maha = gwalk_maha(X_mu, scale, U, Y)
    log_det_cov = log_det_cov_of_cov(COV)
    t0 = time.time()
    const = NDIM * np.log(2*np.pi)
    L = np.exp(-0.5 * (const + log_det_cov[...,None] + maha))
    t1 = time.time()
    print("  Numpy time:\t\t%f seconds!"%(t1-t0))
    return L

######## Tests ########

def test_maha():
    print("Testing mahalanobis distance:")
    maha1 = numpy_maha()
    maha2 = cext_maha()
    assert np.allclose(maha1, maha2)
    print("  cext pass!")

def test_pdf_exp_product():
    print("Testing pdf_exp_product:")
    L1 = numpy_pdf_exp_product()
    L2 = cext_pdf_exp_product()
    assert np.allclose(L1, L2)
    print("  C extension pass!")

######## Main ########
def main():
    test_maha()
    test_pdf_exp_product()
    return

######## Execution ########
if __name__ == "__main__":
    main()



'''Python evaluation of C extensions for multivariate normal distributions'''

######## Imports ########
import numpy as np
from scipy import stats
from ._mahalanobis_distance import _maha
from ._multivariate_normal_pdf_utils import _pdf_exp_product, _log_pdf_exp_product
from ._multivariate_normal_pdf_utils import _multivariate_normal_pdf
from ._multivariate_normal_pdf_utils import _multivariate_normal_log_pdf
from gwalk.multivariate_normal.decomposition import std_of_cov
from gwalk.multivariate_normal.decomposition import cov_rescale
import time

######## Declarations ########

__all__ = [
           "maha",
           "pdf_exp_product",
           "pdf",
          ]

######## Functions ########
def maha(mu, scale, U, sample):
    '''Calculate the mahalanobis distance between a set of points mu and
        many sets of points (sample)

    Parameters
    ----------
    mu : `~numpy.ndarray` (ngauss, ndim)
        The first set of points (traditionally the mean of some set of
            Gaussians
    scale : `~numpy.ndarray` (ngauss, ndim)
        Scale factors for each mu
    U : `~numpy.ndarray` (ngauss, ndim, ndim)
        Some orthonormal eigenvectors describing the space
            we are estimating the distance in;
            traditionally related to the Covariance of a set of Gaussians
    sample : `~numpy.ndarray` (npts, ndim)
        The sample points we are comparing to mu

    Returns
    -------
    maha : `~numpy.ndarray` (ngauss, npts)
        The mahalanobis distance
    '''
    #### Check inputs ####
    ## Check mu ##
    # Check that mu is a numpy array
    if not isinstance(mu, np.ndarray):
        raise TypeError("mu should be a numpy array, but is:", type(mu))
    # Handle the case where mu is 1D
    if len(mu.shape) == 1:
        ngauss = 1
        ndim = mu.shape[0]
        mu = mu.reshape((ngauss, ndim))
    # Handle the case where mu is 2D
    elif len(mu.shape) == 2:
        ngauss, ndim = mu.shape
    # Handle the cases where mu is not the correct shape
    else:
        raise RuntimeError("Unkown mu shape:", mu.shape)
    
    ## Check scale ##
    if scale is None:
        scale = np.ones(ndim)
    # Check that scale is a numpy array
    if not isinstance(scale, np.ndarray):
        raise TypeError("scale should be an ndarray, but is:", type(scale))
    # Handle the case where scale is a 1D array
    if len(scale.shape) == 1:
        # Case ngauss == 1
        if ngauss == 1:
            scale = scale.reshape((ngauss, ndim))
        # Case ngauss == ngauss
        elif scale.size == ndim:
            scale = np.tile(scale, (ngauss, 1))
        else:
            raise RuntimeError("Unknown scale shape %s with mu shape %s"%(str(scale.shape),str(mu.shape)))
    # Handle the case where scale is a 2D array
    elif len(scale.shape) == 2:
        if not ((scale.shape[0] == ngauss) and (scale.shape[1] == ndim)):
            raise RuntimeError("Unknown scale shape %s with mu shape %s"%(str(scale.shape),str(mu.shape)))
    # Handle the case where scale has an unknown shape
    else:
        raise RuntimeError("Unknown scale shape:", scale.shape)

    ## Check U ##
    # Check that U is a numpy array
    if not isinstance(U, np.ndarray):
        raise TypeError("U should be an array, but is type:", type(U))
    # Check the size of U
    if not U.size == ngauss*ndim*ndim:
        raise RuntimeError("There is something wrong with U.")
    # Check the last two indices
    if not ((U.shape[-1] == ndim) and (U.shape[-2] == ndim)):
        raise RuntimeError("U has incompatible shape %s with mu shape: %s"%(str(U.shape), str(mu.shape)))

    ## Check sample ##
    # Check that sample is a numpy array
    if not isinstance(sample, np.ndarray):
        raise TypeError("sample should be an ndarray, but is type:", type(sample))
    # Check the shape of sample
    if not (len(sample.shape) == 2):
        raise RuntimeError("sample shape should have dimension 2; sample shape:", sample.shape)
    # Check the dimensions
    if not (sample.shape[1] == ndim):
        raise RuntimeError("Sample has wrong number of dimensions.")
    # Get npts
    npts = sample.shape[0]

    return _maha(mu, scale, U, sample)

def pdf_exp_product(maha, log_det_cov, ndim, log_scale = False):
    ''' Calculate the value of the multivariate normal pdf given 
        the mahalanobis evaluations and the log_det_cov factor

    Parameters
    ----------
    maha : `~numpy.ndarray` (ngauss, npts)
        The mahalanobis distance
    log_det_cov : `~numpy.ndarray` (ngauss,)
        The log of the determinant of the covariance parameters
    ndim : int
        The number of dimensions of your data
    log_scale : bool
        Calculate the log pdf instead of linear scale?

    Returns
    -------
    pdf : `~numpy.ndarray` (ngauss, npts)
        The pdf of one or more gaussians on a set of points
    '''
    #### Check inputs ####
    ## Check maha ##
    # Check that maha is a numpy array
    if not isinstance(maha, np.ndarray):
        raise TypeError("maha should be a numpy array, but is:", type(maha))
    # This should be 2D
    if len(maha.shape) != 2:
        raise RuntimeError("maha should be a 2D matrix, but has shape:", maha.shape)
    ngauss = maha.shape[0]
    npts = maha.shape[1]

    ## Check log_det_cov ##
    if not isinstance(log_det_cov, np.ndarray):
        raise TypeError("log_det_cov should be a numpy array, but is:", type(log_det_cov))
    # log_det_cov should be 1D
    if not ((len(log_det_cov.shape) == 1) and (log_det_cov.size == ngauss)):
        raise RuntimeError("log_det_cov should have shape (ngauss,), but has shape:", log_det_cov.shape)

    ## Check log_scale ##
    if log_scale:
        return _log_pdf_exp_product(maha, log_det_cov, ndim)
    else:
        return _pdf_exp_product(maha, log_det_cov, ndim)

#### multivariate normal pdf ####
def pdf(
        mu,
        Sigma,
        sample,
        scale = False,
        log_scale = False,
        dynamic_rescale=False,
       ):
    '''Find the likelihood of some data with a particular guess of parameters
    Parameters
    ----------
    mu : `~numpy.ndarray` (ngauss, ndim)
        The first set of points (traditionally the mean of some set of
            Gaussians
    Sigma : `~numpy.ndarray` (ngauss, ndim, ndim)
    sample : `~numpy.ndarray` (npts, ndim)
        The sample points we are comparing to mu
    scale: array like, shape = (ngauss,ndim)
        Input scale for different parameter guesses
            if False: assume input data is PHYSICAL
            if True: assume input data is SCALED
            if (len(Y),) array: scale input by given values
            if (ngauss, len(Y)): Each sample gaussian has its own scale
    log_scale: bool, optional
        Input return log likelihood instead of likelihood?

    Returns
    -------
        L - array of likelihoods (n_gauss, n_pts)
    '''
    #### Check inputs ####
    ## Check mu ##
    # Check that mu is a numpy array
    if not isinstance(mu, np.ndarray):
        raise TypeError("mu should be a numpy array, but is:", type(mu))
    # Handle the case where mu is 1D
    if len(mu.shape) == 1:
        ngauss = 1
        ndim = mu.shape[0]
        mu = mu.reshape((ngauss, ndim))
    # Handle the case where mu is 2D
    elif len(mu.shape) == 2:
        ngauss, ndim = mu.shape
    # Handle the cases where mu is not the correct shape
    else:
        raise RuntimeError("Unkown mu shape:", mu.shape)

    ## Check Sigma ##
    # Check that sigma is a numpy array
    if not isinstance(Sigma, np.ndarray):
        raise TypeError("Sigma should be a numpy array, but is:", type(Sigma))
    # Handle the case where Sigma is 1D
    if len(Sigma.shape) == 1:
        if Sigma.size != ndim:
            raise RuntimeError("Unknown Sigma shape: ",Sigma.shape)
        else:
            std = np.tile(Sigma, (ngauss, 1))
            std_expand = np.tensordot(std, np.ones(ndim), axes=0)
            Sigma = std_expand * np.transpose(std_expand, axes=[0,2,1])
            # TODO this should be handled by a special function
    # Handle the case where Sigma is 2D
    elif len(Sigma.shape) == 2:
        # Handle the case where Sigma is (ndim, ndim)
        if (Sigma.shape[0] == ndim) and (Sigma.shape[1] == ndim):
            # This is a covariance. Tile it
            Sigma = np.tile(Sigma, (ngauss, 1, 1))
        # Handle the case where Sigma is (ngauss, ndim)
        elif (Sigma.shape[0] == ngauss) and (Sigma.shape[1] == ndim):
            # This is an array of stds
            std_expand = np.tensordot(std, np.ones(ndim), axes=0)
            Sigma = std_expand * np.transpose(std_expand, axes=[0,2,1])
            # TODO this should be handled by a special function
        else:
            raise RuntimeError("Unknown Sigma shape:", Sigma.shape)
    # Handle the case where Sigma is 3D
    elif len(Sigma.shape) == 3:
        # Handle the case where Sigma is (ngauss, ndim, ndim)
        if not ((Sigma.shape[0] == ngauss) and (Sigma.shape[1] == ndim) and (Sigma.shape[2] == ndim)):
            raise RuntimeError("Unknown Sigma shape:", Sigma.shape)
    # Handle weird and unknown Sigma shapes
    else:
        raise RuntimeError("Unknown Sigma shape:", Sigma.shape)

    
    ## Check sample ##
    # Check that sample is a numpy array
    if not isinstance(sample, np.ndarray):
        raise TypeError("sample should be an ndarray, but is type:", type(sample))
    # Check the shape of sample
    if not (len(sample.shape) == 2):
        raise RuntimeError("sample shape should have dimension 2; sample shape:", sample.shape)
    # Check the dimensions
    if not (sample.shape[1] == ndim):
        raise RuntimeError("Sample has wrong number of dimensions.")
    # Get npts
    npts = sample.shape[0]

    ## Dynamic rescale ##
    if dynamic_rescale:
        std = std_of_cov(Sigma)
        scale = 1/std

    ## Check scale ##
    # Check that scale is a numpy array
    if ((scale is False) or (scale is None)):
        scale = np.ones(ndim)
    if not isinstance(scale, np.ndarray):
        raise TypeError("scale should be an ndarray, but is:", type(scale))
    # Handle the case where scale is a 1D array
    if len(scale.shape) == 1:
        # Case ngauss == 1
        if ngauss == 1:
            scale = scale.reshape((ngauss, ndim))
        # Case ngauss == ngauss
        elif scale.size == ndim:
            scale = np.tile(scale, (ngauss, 1))
        else:
            raise RuntimeError("Unknown scale shape %s with mu shape %s"%(str(scale.shape),str(mu.shape)))
    # Handle the case where scale is a 2D array
    elif len(scale.shape) == 2:
        if not ((scale.shape[0] == ngauss) and (scale.shape[1] == ndim)):
            raise RuntimeError("Unknown scale shape %s with mu shape %s"%(str(scale.shape),str(mu.shape)))
    # Handle the case where scale has an unknown shape
    else:
        raise RuntimeError("Unknown scale shape:", scale.shape)

    ## Eigenvector Decomposition ##
    # Find eigenvalues and eigenvectors (see scipy)
    Sigma = cov_rescale(Sigma, scale)
    s, u = np.linalg.eigh(Sigma)
    # Find the precision of the datatype used to store eigenvalues
    eps = stats._multivariate._eigvalsh_to_eps(s)
    # Calcualte the multivariate normal distribution
    if log_scale:
        L = _multivariate_normal_log_pdf(mu, scale, s, u, sample, eps) + np.log(np.prod(scale, axis=1)).reshape((ngauss, 1))
    else:
        L = _multivariate_normal_pdf(mu, scale, s, u, sample, eps) * np.prod(scale, axis=1).reshape((ngauss, 1))
    # Flatten L
    if ngauss == 1:
        L = L.flatten()

    return L



'''\
Mesh grids and GPR fits for 1D and 2D marginal intermediary density estimates.
'''

class Mesh(object):
    '''\
    Mesh object
    '''
    def __init__(
                 self,
                 ndim,
                 std,
                 limits,
                 marginals,
                 attrs,
                 evaluation_res = 10,
                ):
        '''Initialize a mesh object

        Parameters
        ----------
        ndim: int
            Input number of dimensions modeled
        std: array like, shape = (ndim,)
            Input sample variance, useful for some things
        limits: array like, shape = (2,ndim)
            Input limits for space of mesh
        marginals: dict
            Input The marginal information, where data and interpolations are stored
        attrs: dict
            Input Attributes associated with the mesh object, See fit
            attrs = {
                     "ndim"         : ndim,
                     "min_bins"     : min_bins,
                     "max_bins1d"   : max_bins1d,
                     "max_bins2d"   : max_bins2d,
                     "ks_threshold" : ks_threshold,
                     "grab_edge"    : grab_edge,
                     "use_cython"   : use_cython,
                     "whitenoise"   : whitenoise,
                     "sparse"       : sparse,
                     "order"        : order,
                    }
        evaluation_res: int, optional
            Input The resolution of the evaluation sets for the mesh
        '''
        # Hold onto these
        self.ndim = ndim
        self.std = std
        self.limits = limits
        self.marginals = marginals
        self.attrs = attrs

        # Generate an evaluation set
        self.generate_evaluation_set(evaluation_res)

    ######## Guided Constructor ########

    @classmethod
    def fit(
            cls,
            sample,
            ndim,
            weights=None,
            limits=None,
            min_bins=5,
            max_bins1d=100,
            max_bins2d=20,
            verbose=False,
            ks_threshold=0.001,
            grab_edge=True,
            use_cython=True,
            whitenoise=0.001,
            sparse=True,
            order=1,
            std_limit_max=4,
           ):
        '''Fit samples and build a mesh object

        Parameters
        ----------
        sample: array like, shape = (npts, ndim)
            Input samples from some density we would like to model
        ndim: int
            Input dimensionality of data
        weights: array like, shape = (npts,), optional
            Input weights for weighted samples
        limits: array like, shape = (2,ndim), optional
            Input limits for sample space
        min_bins: int, optional
            Input minimum histogram bins in a dimension
        max_bins1d: int, optional
            Input maximum 1d histogram bins
        max_bins2d: int, optional
            Input maximum 2d histogram bins
        verbose: bool, optional
            Input give verpose printouts
        ks_threshold: float, optional
            Input ks threshold for fitting marginals with gp_api
        grab_edge: bool, optional
            Input ghost points for histogram edges with interpolation
        use_cython: bool, optional
            Input option to use cython
        whitenoise: float, optional
            Input whitenoise for gaussian process interpolation
        sparse: bool, optional
            Input Sparse cholesky decomposition for gaussian process training?
        order: int, optional
            Input number of times kernel basis functions must be square diff
        std_limit_max: float, optional
            Input Used for setting sensible limits for population
        '''
        ## Imports 
        import time
        import numpy as np
        from gp_api.marginals import Marginal

        # Create attribute dictionary
        attrs = {
                 "ndim"         : ndim,
                 "min_bins"     : min_bins,
                 "max_bins1d"   : max_bins1d,
                 "max_bins2d"   : max_bins2d,
                 "ks_threshold" : ks_threshold,
                 "grab_edge"    : grab_edge,
                 "use_cython"   : use_cython,
                 "whitenoise"   : whitenoise,
                 "sparse"       : sparse,
                 "order"        : order,
                }


        # find the mean
        mean = np.average(sample,weights=weights,axis=0)
        # find the covariance
        cov = np.cov(sample.T, aweights=weights)
        # Find the standard deviations
        std = np.sqrt(np.diag(cov))

        # Identify limits
        if limits is None:
            # Identify soft limits
            soft_limits = np.zeros((ndim, 2))
            for i in range(ndim):
                soft_limits[i,0] = np.min(sample[:,i])
                soft_limits[i,1] = np.max(sample[:,i])
            # Identify maximum std limits
            std_limits = np.zeros((ndim, 2))
            std_limits[:,0] = mean - std_limit_max*std
            std_limits[:,1] = mean + std_limit_max*std
            # Select limits
            limits = np.zeros((ndim,2))
            for i in range(ndim):
                limits[i,0] = max(soft_limits[i,0],std_limits[i,0])
                limits[i,1] = min(soft_limits[i,1],std_limits[i,1])

        # Initialize marginal object
        marginal_object = Marginal(
                                   sample,
                                   limits,
                                   weights=weights,
                                   verbose=verbose,
                                  )

        # Initialize marginals
        marginals = {}
        # Initialize time
        t0 = time.time()

        # Fit 1d marginals
        for i in range(ndim):
            # identify fit tag
            tag = "1d_%d"%i

            # Fit a marginal
            ks, bins, gp_fit, x_train, y_train, y_error = \
                marginal_object.fit_marginal(
                                  indices=[i],
                                  ks_threshold=ks_threshold,
                                  grab_edge=grab_edge,
                                  use_cython=use_cython,
                                  whitenoise=whitenoise,
                                  sparse=sparse,
                                  order=order,
                                  max_bins=max_bins1d,
                                  min_bins=min_bins,
                                )
            # Update dictionary
            marginals["%s_ks"%tag] = ks
            if isinstance(bins,list):
                bins = np.asarray(bins)
            assert isinstance(bins,np.ndarray)
            marginals["%s_bins"%tag] = bins
            marginals["%s_gp_fit"%tag] = gp_fit
            marginals["%s_x_train"%tag] = x_train
            marginals["%s_y_train"%tag] = y_train
            marginals["%s_y_error"%tag] = y_error

        # Fit 2d marginals
        for i in range(ndim):
            for j in range(i):
                # identify fit tag
                tag = "2d_%d_%d"%(i,j)
    
                # Fit a marginal
                ks, bins, gp_fit, x_train, y_train, y_error = \
                    marginal_object.fit_marginal_methods(
                                              [i,j],
                                              ks_threshold=ks_threshold,
                                              grab_edge=grab_edge,
                                              use_cython=use_cython,
                                              whitenoise=whitenoise,
                                              sparse=sparse,
                                              order=order,
                                              min_bins=min_bins,
                                              max_bins=max_bins2d,
                                              mode="search",
                                            )
                # Update dictionary
                marginals["%s_ks"%tag] = ks
                marginals["%s_bins"%tag] = bins
                assert isinstance(bins,np.ndarray)
                marginals["%s_gp_fit"%tag] = gp_fit
                marginals["%s_x_train"%tag] = x_train
                marginals["%s_y_train"%tag] = y_train
                marginals["%s_y_error"%tag] = y_error

        t1 = time.time()

        if verbose:
            print("Marginal dictionary construction time: %f seconds!"%(t1-t0))
            for key in marginals:
                if key.endswith("bins"):
                    tag = key.rstrip("bins").rstrip("_")
                    print(key, marginals[key], marginals["%s_ks"%tag])

        return cls(ndim, std, limits, marginals, attrs)

    ######## Serialization ########

    def save(
             self,
             fname_db,
             label=None,
             compression="gzip",
             **database_kwargs
            ):
        '''Save mesh to file

        Parameters
        ----------
        fname_db: str
            Input file location to save
        label: str, optional
            Input path to group for storing things
        compression: str, optional
            Input hdf5 compression method
        '''
        import numpy as np
        import h5py
        from gwalk.data import Database
        # Load database
        db = Database(fname_db, group=label, **database_kwargs)
        # Set attributes
        db.attr_set_dict('.',self.attrs)
        # Set std
        db.dset_set("std", self.std, compression=compression)
        # Set limits
        db.dset_set("limits", self.limits, compression=compression)
        # Set 1d marginals
        for i in range(self.ndim):
                # Tag to identify fits
                tag = "1d_%d"%i
                # Save bins
                db.dset_set("%s_bins"%tag, np.asarray(self.marginals["%s_bins"%tag]))
                # Save ks
                db.dset_set("%s_ks"%tag, np.asarray(self.marginals["%s_ks"%tag]))
                # Save x_train
                db.dset_set(
                            "%s_x_train"%tag,
                            self.marginals["%s_x_train"%tag],
                            compression=compression,
                           )
                # Save y_train
                db.dset_set(
                            "%s_y_train"%tag,
                            self.marginals["%s_y_train"%tag],
                            compression=compression,
                           )
                # Save y_error
                db.dset_set(
                            "%s_y_error"%tag,
                            self.marginals["%s_y_error"%tag],
                            compression=compression,
                           )

        # Set 2d marginals
        for i in range(self.ndim):
            for j in range(i):
                # Tag to identify fits
                tag = "2d_%d_%d"%(i,j)
                # Save bins
                db.dset_set("%s_bins"%tag, np.asarray(self.marginals["%s_bins"%tag]))
                # Save ks
                db.dset_set("%s_ks"%tag, np.asarray(self.marginals["%s_ks"%tag]))
                # Save x_train
                db.dset_set(
                            "%s_x_train"%tag,
                            self.marginals["%s_x_train"%tag],
                            compression=compression,
                           )
                # Save y_train
                db.dset_set(
                            "%s_y_train"%tag,
                            self.marginals["%s_y_train"%tag],
                            compression=compression,
                           )
                # Save y_error
                db.dset_set(
                            "%s_y_error"%tag,
                            self.marginals["%s_y_error"%tag],
                            compression=compression,
                           )

    @staticmethod
    def exists(fname_db, label=None, **database_kwargs):
        '''Determine if mesh exists
        Parameters
        ----------
        fname_db: str
            Input file location to save
        label: str, optional
            Input path to group for storing things
        '''
        from os.path import isfile
        from gwalk.data import Database
        # Check if the file exists
        if not isfile(fname_db):
            return False
        # Initialize a database
        db = Database(fname_db, **database_kwargs)
        # If no label is given, we are done
        if label is None:
            return True
        # Check if the group exists
        return db.exists(label)


    @classmethod
    def load(cls, fname_db, label=None,**database_kwargs):
        '''Load a mesh from a file
        Parameters
        ----------
        fname_db: str
            Input file location to save
        label: str, optional
            Input path to group for storing things
        '''
        import numpy as np
        import h5py
        from gwalk.data import Database
        from gp_api.utils import fit_compact_nd
        # Load database
        db = Database(fname_db, group=label, **database_kwargs)
        # Load attributes
        attrs = db.attr_dict('.')
        # Load std
        std = db.dset_value("std")
        # Load limits
        limits = db.dset_value("limits")
        # Initialize dictionary for marginals
        marginals = {}

        # Load 1d marginals
        for i in range(attrs["ndim"]):
                # Tag to identify fits
                tag = "1d_%d"%i
                # load bins
                marginals["%s_bins"%tag]    = db.dset_value("%s_bins"%tag)
                # Load ks
                marginals["%s_ks"%tag]      = db.dset_value("%s_ks"%tag)
                # Load x_train
                marginals["%s_x_train"%tag] = db.dset_value("%s_x_train"%tag)
                # load y_train
                marginals["%s_y_train"%tag] = db.dset_value("%s_y_train"%tag)
                # load y_error
                marginals["%s_y_error"%tag] = db.dset_value("%s_y_error"%tag)
                # Fit a gaussian process
                marginals["%s_gp_fit"%tag] = \
                        fit_compact_nd(
                                       marginals["%s_x_train"%tag],
                                       marginals["%s_y_train"%tag],
                                       whitenoise=attrs["whitenoise"],
                                       sparse=attrs["sparse"],
                                       use_cython=attrs["use_cython"],
                                       order=attrs["order"],
                                       train_err=marginals["%s_y_error"%tag],
                                      )

        # Set 2d marginals
        for i in range(attrs["ndim"]):
            for j in range(i):
                # Tag to identify fits
                tag = "2d_%d_%d"%(i,j)
                # load bins
                marginals["%s_bins"%tag]    = db.dset_value("%s_bins"%tag)
                # Load ks
                marginals["%s_ks"%tag]      = db.dset_value("%s_ks"%tag)
                # Load x_train
                marginals["%s_x_train"%tag] = db.dset_value("%s_x_train"%tag)
                # load y_train
                marginals["%s_y_train"%tag] = db.dset_value("%s_y_train"%tag)
                # load y_error
                marginals["%s_y_error"%tag] = db.dset_value("%s_y_error"%tag)
                # Fit a gaussian process
                marginals["%s_gp_fit"%tag] = \
                        fit_compact_nd(
                                       marginals["%s_x_train"%tag],
                                       marginals["%s_y_train"%tag],
                                       whitenoise=attrs["whitenoise"],
                                       sparse=attrs["sparse"],
                                       use_cython=attrs["use_cython"],
                                       order=attrs["order"],
                                       train_err=marginals["%s_y_error"%tag],
                                      )

        return cls(attrs["ndim"], std, limits, marginals, attrs)

    def __eq__(self, other):
        '''\
        General equals method
        '''
        import numpy as np
        # Check for object type
        if not isinstance(other, Mesh):
            return NotImplemented

        # Check dimensionality
        if not self.ndim == other.ndim:
            return False

        # Check attrs
        if not self.attrs == other.attrs:
            return False

        # Check limits
        if not np.allclose(self.limits, other.limits):
            return False

        # Check evaluation res
        if not self.evaluation_res == other.evaluation_res:
            return False

        # Check 1D marginals
        for i in range(self.ndim):
                # Tag to identify fits
                tag = "1d_%d"%i
                # check bins
                if not self.marginals["%s_bins"%tag] == other.marginals["%s_bins"%tag]:
                    return False
                # check ks
                if not np.allclose(
                                   self.marginals["%s_ks"%tag],
                                   other.marginals["%s_ks"%tag],
                                  ):
                    return False
                # check x_train
                if not np.allclose(
                                   self.marginals["%s_x_train"%tag],
                                   other.marginals["%s_x_train"%tag],
                                  ):
                    return False
                # check y_train
                if not np.allclose(
                                   self.marginals["%s_y_train"%tag],
                                   other.marginals["%s_y_train"%tag],
                                  ):
                    return False
                # check y_error
                if not np.allclose(
                                   self.marginals["%s_y_error"%tag],
                                   other.marginals["%s_y_error"%tag],
                                  ):
                    return False

        # Check 2d marginals
        for i in range(self.ndim):
            for j in range(i):
                # Tag to identify fits
                tag = "2d_%d_%d"%(i,j)
                # check bins
                if not self.marginals["%s_bins"%tag].size == other.marginals["%s_bins"%tag].size:
                    return False
                if not np.allclose(
                                   self.marginals["%s_bins"%tag],
                                   other.marginals["%s_bins"%tag],
                                  ):
                    return False
                # check ks
                if not np.allclose(
                                   self.marginals["%s_ks"%tag],
                                   other.marginals["%s_ks"%tag],
                                  ):
                    return False
                # check x_train
                if not np.allclose(
                                   self.marginals["%s_x_train"%tag],
                                   other.marginals["%s_x_train"%tag],
                                  ):
                    return False
                # check y_train
                if not np.allclose(
                                   self.marginals["%s_y_train"%tag],
                                   other.marginals["%s_y_train"%tag],
                                  ):
                    return False
                # check y_error
                if not np.allclose(
                                   self.marginals["%s_y_error"%tag],
                                   other.marginals["%s_y_error"%tag],
                                  ):
                    return False

        return True

    ######## Functions ########

    def generate_evaluation_set(
                                self,
                                res,
                               ):
        '''Generate a set of evaluation points for each marginal

        Parameters
        ----------
        res: int
            Evaluation grid resolution
        '''
        import numpy as np

        # Remember this resolution
        self.evaluation_res = res

        # Generate 1D evaluation set
        for i in range(self.ndim):
                self.generate_1d_evaluations(i,res)

        # Generate 2D evaluation set
        for i in range(self.ndim):
            for j in range(i):
                self.generate_2d_evaluations(i,j,res)

    def generate_1d_evaluations(
                                self,
                                index,
                                res
                               ):
        '''Generate an evaluation set in  the ith dimension

        Parameters
        ----------
        index: int
            Input which dimension would we like to generate an evaluation set for
        res: int
            Input how many points would we like to include on a uniform grid
        '''
        # Imports
        import numpy as np
        from gp_api.utils import sample_hypercube
        # Generate hypercube samples
        grid_samples = sample_hypercube(np.asarray([self.limits[index]]),res)
        x_test = grid_samples
        # Save samples
        self.marginals["1d_%d_x_test"%index] = x_test
        # Generate y samples
        y_test = self.marginals["1d_%d_gp_fit"%index].mean(
            self.marginals["1d_%d_x_test"%index])
        # Fix up y samples
        y_test[y_test < 0.] = 0.
        y_test/= np.sum(y_test)
        # Save y samples
        self.marginals["1d_%d_y_test"%index] = y_test
        lny = np.log(y_test)
        keep = np.isfinite(lny)
        self.marginals["1d_%d_x_test_log"%index] = x_test[keep]
        self.marginals["1d_%d_y_test_log"%index] = y_test[keep]
        self.marginals["1d_%d_lny_test_log"%index] = lny[keep]
        
    def generate_2d_evaluations(
                                self,
                                index,
                                jndex,
                                res
                               ):
        '''Generate an evaluation set in  the ith dimension

        Parameters
        ----------
        index: int
            Input which dimension would we like to generate an evaluation set for
        jndex: int
            Input which other dimension would we like to generate an evaluation set for
        res: int
            Input how many points would we like to include on a uniform grid
        '''
        # Imports
        import numpy as np
        from gp_api.utils import sample_hypercube
        # Generate hypercube samples
        grid_samples = sample_hypercube(
            np.asarray([self.limits[index],self.limits[jndex]]),res)
        x_test = grid_samples
        # Save samples
        self.marginals["2d_%d_%d_x_test"%(index,jndex)] = x_test
        # Generate y samples
        y_test = self.marginals["2d_%d_%d_gp_fit"%(index,jndex)].mean(
            self.marginals["2d_%d_%d_x_test"%(index,jndex)])
        # Fix up y samples
        y_test[y_test < 0.] = 0.
        y_test/= np.sum(y_test)
        # Save y samples
        self.marginals["2d_%d_%d_y_test"%(index,jndex)] = y_test
        lny = np.log(y_test)
        keep = np.isfinite(lny)
        self.marginals["2d_%d_%d_x_test_log"%(index,jndex)] = x_test[keep]
        self.marginals["2d_%d_%d_y_test_log"%(index,jndex)] = y_test[keep]
        self.marginals["2d_%d_%d_lny_test_log"%(index,jndex)] = lny[keep]

    def fetch_1d_evaluations(self, index, log_scale=False):
        '''Return the evaluation set for 1 dimension

        Parameters
        ----------
        index: int
            Input index of marginals you would like
        log_scale: bool, optional
            Input Return log scale of values?
        '''
        if log_scale:
            return \
                self.marginals["1d_%d_x_test_log"%index], \
                self.marginals["1d_%d_y_test_log"%index], \
                self.marginals["1d_%d_lny_test_log"%index]
        else:
            return \
                self.marginals["1d_%d_x_test"%index], \
                self.marginals["1d_%d_y_test"%index]

    def fetch_2d_evaluations(self, index, jndex, log_scale=False):
        '''Return the evaluation set for 2 dimensions

        Parameters
        ----------
        index: int
            Input index of marginals you would like
        jndex: int
            Input second index of marginals you would like
        log_scale: bool, optional
            Input Return log scale of values?
        '''
        if log_scale:
            return \
                self.marginals["2d_%d_%d_x_test_log"%(index,jndex)], \
                self.marginals["2d_%d_%d_y_test_log"%(index,jndex)], \
                self.marginals["2d_%d_%d_lny_test_log"%(index,jndex)]
        else:
            return \
                self.marginals["2d_%d_%d_x_test"%(index,jndex)], \
                self.marginals["2d_%d_%d_y_test"%(index,jndex)]

    def polyfit_mu_sig_1d(self, x, y, sig_default, limits):
        '''\
        Return polynomial coefficients or best guess

        Parameters
        ----------
        x: array like, shape = (npts,)
            Input space values
        y: array like, shape = (npts,)
            Input function values
        sig_default: float
            Input default value of sigma
        limits: array like, shape = (2,)
            Input limits for x space
        '''
        # Imports 
        import numpy as np
        # Call polyfit
        a, b, c = np.polyfit(x,y,2)
        # check if a is viable
        if a < 0:
            # Use polynomial coefficients
            mu = -(0.5*b/a)
            sig = -0.5/a
        else:
            # Use maximum of data
            x = x[1:-1]
            y = y[1:-1]
            mu = x[np.argmax(y)]
            sig = sig_default
        # check limits
        if (mu < limits[0]) or (mu > limits[1]):
            # Use maximum of data
            x = x[1:-1]
            y = y[1:-1]
            mu = x[np.argmax(y)]

        return mu, sig




    ######## MultivariateNormal tools ########

    #### Call Constructor ####
    def construct_nal(
                      self,
                      seed=0,
                      sig_max=None,
                      labels=None,
                     ):
        ''' Construct a bounded multivariate normal model

        Parameters
        ----------
        seed: int, optional
            Input seed for random state
        sig_max: float, optional
            Input maximum sigma parameters, relative to scale
        '''
        # Imports
        import numpy as np
        from gwalk.model.parameter import Parameter
        from gwalk.bounded_multivariate_normal import MultivariateNormal
        # Initialize parameter list
        params = []
        # Loop parameters
        for i in range(self.ndim):
            # Pick a parameter guess
            guess = (self.limits[i][1] + self.limits[i][0])/2

            # Pick a label
            if labels is None:
                label = None
            else:
                label = labels[i]

            # Construct parameter 
            p = Parameter("p_%d"%i,guess,self.limits[i],label)
            params.append(p)

        # Construct Bounded Multivariate Normal object
        MV = MultivariateNormal(
                                params,
                                self.std,
                                seed,
                                sig_max,
                               )
        return MV

    def nal_save_kl(
                    self,
                    MV,
                    fname_nal,
                    label,
                    attrs=None,
                    mode='mean',
                    better='False',
                   ):
        '''Save MV object with kl divergence
        
        Parameters
        ----------
        MV: MultivaraiteNormal object
            Input bounded multivariate normal object
        fname_nal: str
            Input file location for nal fits
        label: str
            Input path to fit group
        attrs: dict, optional
            Input additional attributes to save with nal fit
        better: bool, optional
            Input save fit only if better?
        '''
        # Imports
        from gwalk.bounded_multivariate_normal import MultivariateNormal
        # Initialize attrs if none
        if attrs is None:
            attrs = {}
        # Get kl divergence
        attrs["kl"] = self.nal_kl_div(MV,MV.read_guess(),mode=mode).flatten()
        # Check for better
        if better:
            # Check if the fit already exists
            if MV.exists(fname_nal, label):
                # Load the existing fit
                MVexist = MultivariateNormal.load(fname_nal, label)
                # check MVexist[kl]
                kl_exist = self.nal_kl_div(MVexist,MVexist.read_guess(),mode=mode).flatten()
                # If kl_exist is lower than kl, return
                if kl_exist < attrs["kl"]:
                    return
        # Save fit
        MV.save(fname_nal, label, attrs=attrs)


    #### Convergence ####
    def nal_kl_1d(
                  self,
                  MV,
                  X,
                  index,
                 ):
        ''' Evaluate the 1D kl divergence between a mesh and a normal model

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts,nparam)
            Input Guesses for parameter values for gaussians
        index: integer
            Input dimension we would like to get the kl divergence for
        '''
        from relative_entropy_cython import relative_entropy_alt
        import numpy as np

        ## Prep ##
        X = MV.check_sample(X)
        # Check things
        n_gauss, n_params = X.shape
        ## Calculate 1D kl divergences ##
        # Get mesh inputs
        y_mesh, L_mesh, lnL_mesh = \
            self.fetch_1d_evaluations(index,log_scale=True)
        # Get lnL_norm
        lnL_norm = MV.likelihood(
                                 y_mesh,
                                 X=X,
                                 indices=[index],
                                 log_scale=True,
                                )

        # Identify the parts to keep
        keep = np.isfinite(lnL_norm)
        keep = np.prod(keep,axis=0).astype(bool)
        keep_tile = np.tile(keep,(lnL_norm.shape[0],1))
        # Update the scaling constant
        L_sum = np.sum(L_mesh[keep])
        # Update lnL_norm
        lnL_norm = lnL_norm[keep_tile].reshape(lnL_norm.shape[0],np.sum(keep))
        #print(lnL_mesh[keep])
        #print(lnL_norm)
        kl = relative_entropy_alt(
                                  L_mesh[keep]/L_sum,
                                  lnL_mesh[keep]-np.log(L_sum),
                                  lnL_norm
                                 )
        return kl


    #### Convergence ####
    def nal_kl_2d(
                  self,
                  MV,
                  X,
                  index,
                  jndex,
                 ):
        ''' Evaluate the 1D kl divergence between a mesh and a normal model

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts,nparam)
            Input Guesses for parameter values for gaussians
        index: integer
            Input dimension we would like to get the kl divergence for
        jndex: integer
            Input other dimension we would like the kl divergence for
        '''
        from relative_entropy_cython import relative_entropy_alt
        import numpy as np

        ## Prep ##
        X = MV.check_sample(X)
        # Check things
        n_gauss, n_params = X.shape
        ## Calculate 1D kl divergences ##
        # Get mesh inputs
        y_mesh, L_mesh, lnL_mesh = \
            self.fetch_2d_evaluations(index,jndex,log_scale=True)
        # Get lnL_norm
        lnL_norm = MV.likelihood(
                                 y_mesh,
                                 X=X,
                                 indices=[index,jndex],
                                 log_scale=True,
                                )
        keep = np.isfinite(lnL_norm)
        keep = np.prod(keep,axis=0).astype(bool)
        keep_tile = np.tile(keep,(lnL_norm.shape[0],1))
        # Update the scaling constant
        L_sum = np.sum(L_mesh[keep])
        # Update lnL_norm
        lnL_norm = lnL_norm[keep_tile].reshape(lnL_norm.shape[0],np.sum(keep))
        #print(lnL_mesh[keep])
        #print(lnL_norm)
        kl = relative_entropy_alt(
                                  L_mesh[keep]/L_sum,
                                  lnL_mesh[keep]-np.log(L_sum),
                                  lnL_norm
                                 )
        return kl

    def nal_kl_div(
                   self,
                   MV,
                   X=None,
                   mode="mean",
                  ):
        ''' Calculate the KL Divergence for a set of parameters

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts, nparams),optional
            Input test params for kl divergence
        mode: str
            Input mode of outputs
        '''
        # Imports 
        import numpy as np
        # Use built in value for X
        if X is None:
            X = MV.read_guess()
        # Consider only valid guesses for X
        X = MV.check_sample(X)
        n_gauss = X.shape[0]

        ## Calculate kl divergences ##
        kl1d = np.zeros((n_gauss,self.ndim))
        for i in range(self.ndim):
                kl1d[:,i] = self.nal_kl_1d(MV,X,i,)
        kl2d = np.zeros((n_gauss,self.ndim,self.ndim))
        for i in range(self.ndim):
            for j in range(i):
                kl2d[:,i,j] = self.nal_kl_2d(MV,X,i,j)

        # Calculate the goodness of fit statistic
        if mode == "sum":
            kl = np.sum(kl1d,axis=-1) + np.sum(np.sum(kl2d,axis=-1),axis=-1)
        elif mode == "rms":
            kl = np.sqrt(np.sum(kl1d**2,axis=-1) + np.sum(np.sum(kl2d**2,axis=-1),axis=-1))
        elif mode == "mean":
            kl1d_sum = np.sum(kl1d,axis=-1)
            kl2d_sum = np.sum(np.sum(kl2d,axis=-1),axis=-1)
            kl = 0.5 * (
                        (kl1d_sum/self.ndim) +\
                        (kl2d_sum/(self.ndim*(self.ndim -1)//2))
                       )
            kl_ind = np.argmin(kl)
        elif mode == "component":
            kl = np.empty((X.shape[0],len(MV._parameters)-self.ndim))
            # First consider the 1D kl divergences
            for i in range(self.ndim):
                kl[:,MV.p_map["mu_%d"%i]-1] = kl1d[:,i]
            # The 2d kl divergences have a bearing on 5 different parameters
            for i in range(self.ndim):
                for j in range(i):
                    kl[:,MV.p_map["cor_%d_%d"%(i,j)]-self.ndim-1] = kl2d[:,i,j]
        elif mode == "parameter":
            kl = np.zeros((X.shape[0],len(MV._parameters)))
            # First consider the 1D kl divergences
            for i in range(self.ndim):
                kl[:,MV.p_map["mu_%d"%i]] = kl1d[:,i]
                kl[:,MV.p_map["std_%d"%i]] = kl1d[:,i]
            # The 2d kl divergences have a bearing on 5 different parameters
            for i in range(self.ndim):
                for j in range(i):
                    kl[:,MV.p_map["mu_%d"%i]] += kl2d[:,i,j]
                    kl[:,MV.p_map["std_%d"%i]] += kl2d[:,i,j]
                    kl[:,MV.p_map["mu_%d"%j]] += kl2d[:,i,j]
                    kl[:,MV.p_map["std_%d"%j]] += kl2d[:,i,j]
                    kl[:,MV.p_map["cor_%d_%d"%(i,j)]] = kl2d[:,i,j]
            # for each parameter, we want to use the average of the
            # kl divergences associated with that parameter.
            kl[:,:self.ndim] /= 3
            kl[:,self.ndim:2*self.ndim] /= 3
        else:
            raise RuntimeError("Unknown kl mode: %s"%(mode))

        return kl

    def nal_kl_function(MV,mode="mean",kl_sensitivity=None,):
        ''' Return a function which evaluates the kl divergence given params

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        X: array like, shape = (npts, nparams),optional
            Input test params for kl divergence
        mode: str
            Input mode of outputs
        kl_sensitivity: bool
            Dilute kl divergences with some small factor?
        '''
        def kl_div(X):
            kl = self.nal_kl_div(MV,X=X,mode=mode,kl_sensitivity=kl_sensitivity)
            return kl
        return kl_div

    #### Guessing ####
    def nal_mesh_guesses(
                          self,
                          MV,
                         ):
        ''' Fit the bounded multivariate normal model to the mesh
        
        Parameters
        ----------
        MV: MultivariateNormal object
            Input some initialized Multivariate Normal object
        '''
        # Imports 
        import numpy as np
        # Identify number of guesses
        n_guess = 3
        # Initialize guesses
        X = MV.read_guess()
        Xg = np.tile(X,(n_guess,1))

        # Generate 1D evaluation guesses
        for i in range(self.ndim):
            # Load values
            y_test, L_test = self.fetch_1d_evaluations(i)
            # adjust coordinates
            keep = L_test > 0
            y_test = y_test[keep].flatten()/MV.scale[i]
            L_test = np.log(L_test[keep].flatten())
            # Find maximum
            mu, sig = self.polyfit_mu_sig_1d(y_test, L_test, X[i+self.ndim], self.limits[i])
            Xg[0,i+1] = mu
            Xg[0,i+self.ndim+1] = sig

        # Generate 1D training guesses
        for i in range(self.ndim):
            # Load values
            y_train = self.marginals["1d_%d_x_train"%i]
            L_train = self.marginals["1d_%d_y_train"%i]
            bins = int(self.marginals["1d_%d_bins"%i])

            # Rescale training set 1
            y_train_1 = y_train[:bins].flatten()/MV.scale[i]
            L_train_1 = L_train[:bins].flatten()
            keep_1 = L_train_1 > 0
            y_train_1 = y_train_1[keep_1]
            L_train_1 = np.log(L_train_1[keep_1])
            # Rescale training set 2
            y_train_2 = y_train[bins:].flatten()/MV.scale[i]
            L_train_2 = L_train[bins:].flatten()
            keep_2 = L_train_2 > 0
            y_train_2 = y_train_2[keep_2]
            L_train_2 = np.log(L_train_2[keep_2])

            # fit training set 1
            mu, sig = self.polyfit_mu_sig_1d(y_train_1, L_train_1, X[i+self.ndim], self.limits[i])
            Xg[1,i+1] = mu
            Xg[1,i+self.ndim+1] = sig

            # fit training set 2
            mu, sig = self.polyfit_mu_sig_1d(y_train_2, L_train_2, X[i+self.ndim], self.limits[i])
            Xg[2,i+1] = mu
            Xg[2,i+self.ndim+1] = sig

        return Xg

    #### Initialization ####
    def nal_init_walkers(
                         self,
                         MV,
                         nwalk,
                         Xg=None,
                         f_opt=None,
                         f_opt_param=None,
                         sig_multiplier=3,
                         sig_min=1e-5,
                        ):
        '''Initialize random walkers for bounded normal fit optimization

        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        nwalk: int
            Input number of random walkers to initialize
        Xg: array like, shape = (npts, nparams), optional
            Input parameters for guesses
        f_opt: function, optional
            Input likelihood function for likelihood evaluation
        f_opt: function, optional
            Input likelihood function for parametric likelihood evaluation
        sig_multiplier: float, optional
            Input used for generating initial guesses
        '''
        # Imports 
        import numpy as np
        
        ## Check inputs ##
        # Initialize optimization function
        if f_opt is None:
            def f_opt(X):
                L = np.zeros(X.shape[0])
                k = MV.satisfies_constraints(X)
                L[k] = np.power(self.nal_kl_div(MV,X=X[k]),-1)
                L[~k] = 0.
                return L

        if f_opt_param is None:
            def f_opt_param(X):
                L = np.ones(X.shape)
                k = MV.satisfies_constraints(X)
                L[k,1:] = np.power(self.nal_kl_div(MV,X=X[k],mode='parameter')[:,1:],-1)
                L[~k] = 0.
                return L

        # Initialize guesses
        if Xg is None:
            Xg = self.nal_mesh_guesses(MV)
        Xg = MV.check_sample(Xg)
        # Determine guess goodness
        Lg = f_opt_param(Xg)
        keep = np.prod(Lg,axis=-1) > 0.
        # Downselect guesses
        Xg = Xg[keep]
        Lg = Lg[keep]

        # Determine mixmatch guess
        Xm = np.empty(Xg.shape[1])
        for i in range(Xg.shape[1]):
            Xm[i] = Xg[np.argmax(Lg[:,i])][i]
        Xm = MV.check_sample(Xm)
        if bool(MV.satisfies_constraints(Xm)):
            Xg = np.append(Xg,Xm,axis=0)

        # Determine number of guesses
        nguess = Xg.shape[0]

        # Sort guesses
        sort_index = np.argsort(np.prod(Lg,axis=-1))[::-1]
        Xg = Xg[sort_index]
        Lg = Lg[sort_index]

        # If we have more guesses than we need, return the number we need
        if nguess >= nwalk:
            Xg = Xg[:nwalk]
            return Xg

        # If we have less guesses than we need, generate new guesses
        while nguess < nwalk:
            # Generate new random guesses
            mu = np.average(Xg,axis=0)
            sig = sig_multiplier*np.std(Xg[:,1:],axis=0)
            sig[sig < sig_min] = sig_min
            Xn = np.tile(mu,(nwalk-nguess,1))
            Xn[:,1:] += MV.rs.randn(nwalk-nguess,sig.size)*sig
            #Xn = MV.sample_uniform_unconstrained(nwalk - nguess)
            k = MV.satisfies_constraints(Xn)
            Xk = Xn[k]
            Xg = np.append(Xg,Xk,axis=0)
            nguess = Xg.shape[0]

        return Xg

    #### Fit methods ####

    def nal_fit_to_samples(self,MV,sample,weights=None,**kwargs):
        ''' Fit the bounded multivariate normal model to some samples
        Parameters
        ----------
        MV: MultivariateNormal object
            Input some initialized Multivariate Normal object
        sample: array like, shape = (npts, ndim)
            Input Sample array
        weights: array like, shape = (npts,)
            Input weights for sample array
        '''
        # Fit to samples
        MV.fit_simple(sample,w=weights,assign=True)
        return MV

    def nal_genetic_step(
                         self,
                         MV,
                         cur,
                         f_opt,
                         f_opt_param,
                         nwalk,
                         carryover = 0.03,
                         sig_factor = 1.0,
                        ):
        '''Draw a new step randomly within bounds, compare the likelihood
        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        cur: array like, shape = (npts, nparams), optional
            Input parameters for guesses
        f_opt: function
            Input likelihood function for likelihood evaluation
        f_opt: function
            Input likelihood function for parametric likelihood evaluation
        nwalk: int
            Input number of random walkers to initialize
        carryover: float, optional
            Input carryover fraction for genetic algorithm
        sig_factor: float, optional
            Input number of sigma to vary new guesses by
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal

        # Use variance of guesses to determine jump scale
        sig = np.std(cur, axis=0)[1:]

        # Generate parameter likelihood
        Lcur_p = f_opt_param(cur)
        Lcur = np.prod(Lcur_p[:,1:],axis=1)

        # Identify best guesses
        n_carry = int(carryover*nwalk)
        if n_carry > np.sum(Lcur > 0):
            n_carry = np.sum(Lcur > 0)
        carry_index = np.argsort(Lcur)[-n_carry:]
        carry = cur[carry_index]
        Lcarry = Lcur[carry_index]

        ## Breeding ##
        # The breeding pool excludes candidates with fitness zero
        keep = Lcur > 0
        Xb = cur[keep].copy()
        Lb = Lcur[keep]
        Lb_p = Lcur_p[keep]
        Lb /= np.sum(Lb)
        # Pick random parents
        p1 = MV.rs.choice(np.arange(Xb.shape[0]),size=nwalk,p=Lb)
        p2 = MV.rs.choice(np.arange(Xb.shape[0]),size=nwalk,p=Lb)
        for i in range(nwalk):
            # Choose random parameter values
            choices = np.zeros(cur.shape[1],dtype=bool)
            Lb_p_sum = Lb_p[p1[i]] + Lb_p[p2[i]]
            for j in range(1,cur.shape[1]):
                choices[j] = MV.rs.choice(
                       [True,False],
                       p=[
                          Lb_p[p1[i]][j]/Lb_p_sum[j],
                          Lb_p[p2[i]][j]/Lb_p_sum[j],
                         ]
                      )
            cur[i,choices] =  cur[p1][i,choices]
            cur[i,~choices] = cur[p2][i,~choices]
        # Re-evaluate likelihood
        Lcur = f_opt(cur)

        # Hold best guesses over through breeding
        drop_index = np.argsort(Lcur)[:n_carry]
        cur[drop_index] = carry
        Lcur[drop_index] = Lcarry
        # Identify new best guesses
        carry_index = np.argsort(Lcur)[-n_carry:]
        carry = cur[carry_index]
        Lcarry = Lcur[carry_index]

        ## Generate new steps ##
        # loop through each random walker
        new = np.copy(cur)
        new[:,1:] += MV.rs.randn(nwalk,sig.size)*sig*sig_factor
        keep = MV.satisfies_constraints(new)
        new[~keep] = cur[~keep]

        # Determine the likelihood of the new guess
        Lnew = f_opt(new)

        # Determine alpha
        alpha = Lnew/Lcur

        # Decide if to jump
        jumpseed = (MV.rs.uniform(size=nwalk) > (1 - alpha)).astype(bool)
        jumpseed[Lcur==0] = True

        # Jump
        new[~jumpseed] = cur[~jumpseed]
        Lnew[~jumpseed] = Lcur[~jumpseed]

        # Hold best guesses
        drop_index = np.argsort(Lnew)[:n_carry]
        new[drop_index] = carry
        Lnew[drop_index] = Lcarry

        return new.copy(), Lnew.copy()

    #### Random Walk Algorithms ####

    def nal_fit_random_walk(
                            self,
                            MV,
                            cur,
                            f_opt = None,
                            f_opt_param = None,
                            nwalk=100,
                            nstep=100,
                            carryover=0.03,
                            sig_factor=1.0,
                           ):
        '''\
        Begin using a random walk to find the MLE value for our model
        Parameters
        ----------
        MV: MultivariateNormal object
            Input bounded multivariate normal object
        cur: array like, shape = (npts, nparams), optional
            Input parameters for guesses
        f_opt: function, optional
            Input likelihood function for likelihood evaluation
        f_opt: function, optional
            Input likelihood function for parametric likelihood evaluation
        nwalk: int, optional
            Input number of random walkers to initialize
        nstep: int, optional
            Input number of steps for random walkers
        carryover: float, optional
            Input carryover fraction for genetic algorithm
        sig_factor: float, optional
            Input number of sigma to vary new guesses by
        '''
        ## Imports ##
        # Public
        import time
        import numpy as np
        ## Check inputs ##
        # Initialize optimization function
        if f_opt is None:
            def f_opt(X):
                L = np.zeros(X.shape[0])
                k = MV.satisfies_constraints(X)
                L[k] = np.power(self.nal_kl_div(MV,X=X[k]),-1)
                L[~k] = 0.
                return L

        if f_opt_param is None:
            def f_opt_param(X):
                L = np.zeros(X.shape)
                k = MV.satisfies_constraints(X)
                L[k] = np.power(self.nal_kl_div(MV,X=X[k],mode='parameter'),-1)
                L[~k] = 0.
                return L

        # Initialize the best fit
        Lcur = f_opt(cur)
        index = np.argmax(Lcur)
        best_guess = cur[index].copy()
        Lbest = Lcur[index]

        # Do the fit
        for i in range(nstep):
            cur, Lcur = \
                self.nal_genetic_step(
                                      MV,
                                      cur,
                                      f_opt,
                                      f_opt_param,
                                      nwalk,
                                      carryover=carryover,
                                      sig_factor=sig_factor,
                                     )
            # Testing
            if np.max(Lcur) > Lbest:
                j = np.argmax(Lcur)
                if MV.satisfies_constraints(cur[j,:]):
                    best_guess = cur[j,:].copy()
                    Lbest = Lcur[j].copy()

        # Assign the best guess!
        MV.assign_guess(best_guess)#, force=True)

        return




#!/usr/bin/env python3
'''\
The Multivariate Normal object for fitting

Provide infrastructure for models and hyperparameters
'''
#### Module imports ####
from gwalk.data import Database
from gwalk.model import Parameter, Relation, Covariance
from gwalk.utils.multivariate_normal import cov_of_std_cor
from gwalk.utils.multivariate_normal import std_of_cov
from gwalk.utils.multivariate_normal import cor_of_cov
from gwalk.utils.multivariate_normal import mu_of_params
from gwalk.utils.multivariate_normal import std_of_params
from gwalk.utils.multivariate_normal import cor_of_params
from gwalk.utils.multivariate_normal import cov_of_params
from gwalk.utils.multivariate_normal import mu_cov_of_params
from gwalk.utils.multivariate_normal import params_of_norm_mu_cov
from gwalk.utils.multivariate_normal import multivariate_normal_pdf
from gwalk.utils.multivariate_normal import n_param_of_ndim
from gwalk.utils.multivariate_normal import ndim_of_n_param
from gwalk.utils.multivariate_normal import params_reduce_1d
from gwalk.utils.multivariate_normal import params_reduce_2d
from gwalk.utils.multivariate_normal import params_reduce_dd
from gwalk.utils.multivariate_normal import multivariate_normal_marginal1d
from gwalk.utils.multivariate_normal import multivariate_normal_marginal2d
from gwalk.utils.multivariate_normal import analytic_kl_of_params
from gwalk.utils.multivariate_normal import analytic_enclosed_integral

#### Public Imports ####
import warnings
import numpy as np
#### Globals ####
_INV_RT2 = 1./np.sqrt(2)
_LOG_2PI = np.log(2*np.pi)
DEFAULT_SIG_MAX = 3.
DEFAULT_NORM_MAX = 10.

######## Multivariable Normal Object ########
class MultivariateNormal(Relation):
    '''\
    fit data to a Gaussian and provide different methods
    '''
    #### Initialize ####
    def __init__(
                 self,
                 variables,
                 scale,
                 seed=None,
                 sig_max = None,
                 norm = None,
                ):
        '''Initialize a Multivariate Normal Relation

        Parameters
        ----------
        variables: list
        Input A list, tuple, or dictionary of parameter objects
            This does not include the covariance matrix
            which we will generate automatically
            to reduce pain
            Variables specifically describe the domain
                of the function we are fitting.
        scale: array like, shape = (ndim,)
            Input scale for eigenvalue safety
        seed: int, optional
            Input seed for random number generator
        sig_max: float, optional
            Input sets limit on variance parameters
        fit_norm: boolean, optional
            Input decide if to fit a normalization parameter
        max_norm: float, optional
            Input decide the maximum value for the normalization parameter
        '''
        # Check random state
        self.rs = np.random.RandomState(seed)
        self.seed = seed

        # Check sig_max
        if sig_max is None:
            sig_max = DEFAULT_SIG_MAX
        self.sig_max = sig_max

        # Identify the number of dimensions in the domain
        self.ndim = len(variables)

        # Generate the covariance parameters
        cov_parameters = Covariance(
                                    self.ndim,
                                    sig_max=sig_max,
                                    random_state=self.rs,
                                   )
        # Keep track of the scale
        self.scale = scale*np.ones(self.ndim)

        # Generate a temporary relationship with the variables
        temp = Relation(variables,random_state=self.rs)

        # Initialize this object as a "Relation"
        super().__init__(random_state=self.rs)

        # Create a normalization relation
        norm_parameters = Relation(random_state=self.rs)
        if norm is None:
            p = Parameter("norm", 0.0, [0., DEFAULT_NORM_MAX], label="norm")
            norm_parameters.add_parameter(p)
        else:
            norm_parameters.add_parameter(norm)


        # Inialize a Relation object with the mu variables
        mu_parameters = Relation(random_state=self.rs)

        # Rename mu variables
        for i, key in enumerate(temp._parameters):
            p = Parameter(
                          "mu_%d"%(i),
                          temp._parameters[key].guess/self.scale[i],
                          temp._parameters[key].limits/self.scale[i],
                          label=temp._parameters[key].label
                         )
            mu_parameters.add_parameter(p)

        # Contain the entire list of covariance parameters
        temp_obj = norm_parameters + mu_parameters + cov_parameters
        self.__dict__.update(temp_obj.__dict__)

        # Carry functions that identify things about the covariance matrix
        self.read_std      = cov_parameters.read_std
        self.read_cor      = cov_parameters.read_cor
        self.read_cov      = cov_parameters.read_cov
        self.params_of_cov  = cov_parameters.params_of_cov
        self.mu_hypercube  = mu_parameters.hypercube
        self.sample_uniform_variable = mu_parameters.sample_uniform


    #### Tools for working with this object ####
    ## Guess reading tools ##
    def read_norm(self):
        ''' Read the normalization constant'''
        return self._parameters["norm"].guess

    def read_mu(self):
        ''' Read the mu guesses'''
        ## Imports ##
        # Public 
        import numpy as np
        # Initialize mu
        mu = np.zeros(self.ndim)
        # Loop through std parameters
        for i in range(self.ndim):
            # Generate tag
            tag = "mu_%d"%(i)
            # Load value
            mu[i] = self._parameters[tag].guess
        return mu

    def read_scaled(self):
        '''Read the guess to mu and cov'''
        mu = self.read_mu()
        cov = self.read_cov()
        return mu, cov

    def read_physical(self):
        '''Read the guess to mu and cov'''
        mu = self.read_mu()*self.scale
        cov = self.read_cov()*np.outer(self.scale,self.scale)
        return mu, cov

    def read_physical_limits(self):
        '''Read the limits for the physical parameters'''
        return self.summands[1].read_limits()*self.scale[:,None]

    ## Conversions ##

    def mu_of_params(self,X):
        '''Convert parameters to mu values

        Parameters
        ----------
        X: array like, shape = (npts, nparams)
            Input Array of parameter guesses
        '''
        return mu_of_params(X)

    def std_of_params(self,X):
        '''Convert parameters to std values

        Parameters
        ----------
        X: array like, shape = (npts, nparams)
            Input Array of parameter guesses
        '''
        return std_of_params(X)

    def cor_of_params(self, X):
        '''Convert parameters to corelation values

        Parameters
        ----------
        X: array like, shape = (npts, nparams)
            Input Array of parameter guesses
        '''
        return cor_of_params(X)

    def cov_of_params(self, X):
        '''Convert parameters to covariance values

        Parameters
        ----------
        X: array like, shape = (npts, nparams)
            Input Array of parameter guesses
        '''
        return cov_of_params(X)

    def mu_cov_of_params(X):
        '''Convert parameter values to mu and cov values

        Parameters
        ----------
        X: array like, shape = (npts, nparams)
            Input Array of parameter guesses
        '''
        return mu_cov_of_params(X)

    def params_of_norm_mu_cov(mu,cov):
        '''Convert parameter values to mu and cov values

        Parameters
        ----------
        mu: array like, shape = (npts, ndim)
            Input Array of mu values
        cov: array like, shape = (npts, ndim, ndim)
            Input Array of cov values
        '''
        return params_of_norm_mu_cov(mu,cov)

    ## Guess assignment ##

    def satisfies_constraints(self, Xs):
        '''Pick only valid guesses for params

        Parameters
        ----------
        X: array like, shape = (npts, nparams)
            Input Array of parameter guesses
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal
        # Fix data
        Xs = self.check_sample(Xs)
        # Check mu parameters
        c0 = self.summands[0].satisfies_constraints(Xs[:,0])
        # Check mu parameters
        c1 = self.summands[1].satisfies_constraints(Xs[:,1:self.ndim+1])
        # Check covariance
        c2 = self.summands[2].satisfies_constraints(Xs[:,self.ndim+1:])
        constrained_indx = (c0 & c1 & c2)
        # Print statement for errors
        #if (constrained_indx.size == 1) and (not constrained_indx):
        #    print(c1, c2, c3)

        return constrained_indx
                
    def likelihood(
                   self,
                   Y,
                   X=None,
                   w=None,
                   return_prod=False,
                   log_scale=False,
                   scale=False,
                   indices=None,
                   limits=None,
                   lnL_offset=None,
                  ):
        '''Find the likelihood of some data with a particular guess of parameters
        
        Parameters
        ----------
        Y: array like, shape = (npts, ndim)
            Input Array of samples to be evaluated
        X: array like, shape = (ngauss, nparams), optional
            Input Array of parameter guesses
        w: array_like, shape = (npts,), optional
            Input Array of weights for sum. 
                Don't use these unless you know what you are doing
        return_prod: bool, optional
            Input return product of likelihood?
        log_scale: bool, optional
            Input return log likelihood instead of likelihood?
        scale: array like, shape = (ngauss,ndim)
            Input scale for different parameter guesses
                if False: assume input data is PHYSICAL
                if True: assume input data is SCALED
                if (len(Y),) array: scale input by given values
                if (ngauss, len(Y)): Each sample gaussian has its own scale
        indices: list, optional
            Input sometimes we only want to evaluate this on a few dimensions
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal
        from scipy import stats

        # Protect scope
        Y = Y.copy()

        ## Check X ##
        # If X is None, guess is already assigned
        if X is None:
            X = self.read_guess()
        # Check the sample
        X = self.check_sample(X)

        # Read information
        n_gauss, n_param = X.shape
        ndim_X = ndim_of_n_param(n_param)
        assert ndim_X == self.ndim

        # Check indices
        if indices is None:
            indices = np.arange(ndim_X)
        else:
            indices = np.asarray(indices)
        ndim_Y = indices.size

        ## Check Y ##
        # Check is numpy array
        if not isinstance(Y, np.ndarray):
            Y = np.asarray(Y)
        n_sample = Y.shape[0]

        # Check dimensions
        Y = Y.reshape((n_sample, ndim_Y))
            
        ## Check scale ##
        if scale is False:
            scale = np.ones((n_gauss,ndim_X))*self.scale
        elif scale is True:
            scale = np.ones((n_gauss,ndim_X))
        elif np.asarray(scale).size == ndim_X:
            scale = np.ones((n_gauss,ndim_X))*scale
        elif np.asarray(scale).shape == (n_gauss,ndim_X):
            # This is intended behavior
            pass
        else:
            raise RuntimeError("Strange Scale. Aborting.")

        # Isolate indices
        X = params_reduce_dd(X, indices)
        scale = scale[:,indices]
      
        # Evaluate normal pdf
        L = multivariate_normal_pdf(X, Y, scale=scale, log_scale=log_scale)

        # Consider bounds
        if limits is None:
            limits = self.read_physical_limits()
        # If only one set of limits exists, this is easy
        if len(limits.shape) == 2:
            # Check for sensible limits
            assert (limits.shape[0] == self.ndim) and (limits.shape[1] == 2)
            limits = limits[indices,:]
            # generate the keep index
            Y_keep = np.ones((Y.shape[0]),dtype=bool)
            for i in range(len(indices)):
                Y_keep &= Y[:,i] >= limits[i,0]
                Y_keep &= Y[:,i] <= limits[i,1]
            if log_scale:
                L[:,~Y_keep] = -np.inf
            else:
                L[:,~Y_keep] = 0.

        elif len(limits.shape) == 3:
            # Check for sensible limits
            assert limits.shape[0] == n_gauss
            assert limits.shape[1] == self.ndim
            assert limits.shape[2] == 2
            limits = limits[:,indices,:]
            Y_keep = np.ones_like(L,dtype=bool)
            for i in range(n_gauss):
                for j in range(len(indices)):
                    Y_keep[i] &= Y[:,j] >= limits[i,j,0]
                    Y_keep[i] &= Y[:,j] <= limits[i,j,1]
            if log_scale:
                L[~Y_keep] = -np.inf
            else:
                L[~Y_keep] = 0

        # Apply weights
        if not (w is None):
            raise Exception("What is this used for?")
            if X is None:
                L = np.atleast_2d(L.dot(w))
            elif (len(X.shape) == 2) and (X.shape[1] == len(self._parameters)):
                L *= w 
                
        # Check how we want the output
        if return_prod:
            if log_scale:
                L = np.sum(L, axis=1)
            else:
                L = np.prod(L, axis=1)
        if lnL_offset is None:
            lnL_offset = self.read_norm()
        if log_scale:
            L += lnL_offset
        else:
            L *= np.exp(lnL_offset)

        return L

    def pdf(self,Y):
        '''Return the pdf using the current guess for the likelihood
        Parameters
        ----------
        Y: array like, shape = (npts, ndim)
            Input Array of samples to be evaluated
        '''
        return self.likelihood(Y)

    def lnL(self,Y,w=None):
        '''Return the sum of the log of the likelihood
        Parameters
        ----------
        Y: array like, shape = (npts, ndim)
            Input Array of samples to be evaluated
        w: array_like, shape = (npts,), optional
            Input Array of weights for sum. 
                Don't use these unless you know what you are doing
        '''
        return self.likelihood(Y,log_scale=True,return_prod=True,w=w)

    def analytic_kl(self, X2, X1=None, scale1=None, scale2=None):
        '''Return the analytic kl divergence between two Gaussians

        Parameters
        ----------
        X1: array like, shape = (ngauss, nparams), optional
            Input array of gaussian parameters for P distribution
            Defaults to current guess
        X2: array like, shape = (ngauss, nparams)
            Input array of gaussian parameters for Q distribution
        scale1: array like, shape = (ngauss, ndim), optional
            Input scale factors for P distribution
            Defaults to current scale
        scale2: array like, shape = (ngauss, ndim), optional
            Input scale factors for Q distribution
            Defaults to current scale
        '''
        if isinstance(X2, MultivariateNormal):
            other = X2
            X2 = other.read_guess()
            scale2 = np.copy(other.scale)
        if X1 is None:
            X1 = self.read_guess()
        if len(X1.shape) == 1:
            X1 = X1[None,:]
        if len(X2.shape) == 1:
            X2 = X2[None,:]
        if scale1 is None:
            scale1 = self.scale[None,:]
        else:
            if len(scale1.shape) == 1:
                scale1 = scale1[None,:]
            if X1.shape[0] != scale1.shape[0]:
                print("Inconsistent number of guesses for X1")
        if scale2 is None:
            scale2 = self.scale[None,:]
        else:
            if len(scale2.shape) == 1:
                scale2 = scale2[None,:]
            if X2.shape[0] != scale2.shape[0]:
                print("Inconsistent number of guesses for X2")
        if X1.shape[0] > 1:
            raise NotImplementedError
        if X2.shape[0] > 1:
            raise NotImplementedError
        return analytic_kl_of_params(X1, X2, scale1=scale1, scale2=scale2)

    def analytic_enclosed_integral(self,X=None,limits=None,scale=None,assign=False):
        '''Return the analytic enclosed integral for NAL truncation

        Parameters
        ----------
        X: array like, shape = (ngauss, nparams), optional
            Input array of gaussian parameters
            Defaults to current guess
        limits: array like, shape = (ngauss, ndim, 2), optional
            Input array of truncation limits
            Defaults to current limits
        scale: array like, shape = (ngauss, ndim), optional
            Input scale factors 
            Defaults to current scale
        assign: bool, optional
            Input update lnL offset?
        '''
        if X is None:
            self_X = True
            X = self.read_guess()
        else:
            self_X = False
        if limits is None:
            limits = self.read_physical_limits()
        if scale is None:
            scale = np.copy(self.scale[None,:])
        # Calculate integral
        I = analytic_enclosed_integral(X,limits,scale=scale)
        # Assign values
        if assign:
            if self_X:
                assert I.size == 1
                lnL_offset = - np.log(I)
                X[0] = lnL_offset
                self.assign_guess(X)
            else:
                raise RuntimeError("Assign only valid for MV's own params")
        return I

    def normalize(self):
        '''Update the analytic enclosed integral estimate of lnL_offset'''
        self.analytic_enclosed_integral(assign=True)

    #### Resample ####

    def sample_normal_unconstrained(self, rv, size):
        '''Draw a number, size, of random samples
        Parameters
        ----------
        rv: multivariate_normal object
            Input scipy rvs evaluation
        size: int
            Input number of samples to draw
        '''
        ## Imports ##
        # Public
        import numpy as np
        # Generate samples
        samples = np.atleast_2d(rv.rvs(size = int(size),random_state=self.rs))
        if not samples.shape[1] == self.ndim:
            raise RuntimeError("normal samples are the wrong shape")
        return samples

    def sample_normal(self, size, scale=False, params=None):
        '''Draw a number, size, of random samples from inside limits

        This introduces bias, so be careful

        Parameters
        ----------
        size: int
            Input number of samples to draw
        scale: array like, shape = (1,ndim)
            Input scale for different parameter guesses
                if False: assume output data is PHYSICAL
                if True: assume output data is SCALED
        params: array like, shape = (1, nparams), optional
            Input Array of parameter guesses
        '''
        ## Imports ##
        # Public
        import numpy as np
        from scipy.stats import multivariate_normal

        # Initialize samples
        sample_shape = (size, self.ndim)
        Xk = np.zeros((size, self.ndim))

        # Initialize the number of samples we have
        n_keep = 0

        # load params
        if params is None:
            mu = self.read_mu()
            cov = self.read_cov()
        else:
            mu, cov = mu_cov_of_params(params)
            mu = mu[0]
            cov = cov[0]

        # Initialize rv
        try:
            rv = multivariate_normal(mu, cov)
        except RuntimeError as exp:
            from numpy.linalg import eigvals, eigvalsh
            print("Sampling failed")
            print("Satisfies constraints: ")
            print(self.satisfies_constraints(params))
            print("scale: ")
            print(self.scale)
            print("std_dev: ")
            print(std_of_cov(cov))
            print("cor: ")
            print(cor_of_cov(cov))
            print("eig: ")
            print(eigvalsh(cov))
            raise exp

        # Continue doing this until we have the right number of samples
        while n_keep < size:
            # How many samples do we still need?
            n_need = size - n_keep
            # Draw that many samples
            Xs = self.sample_normal_unconstrained(rv, n_need)
            # Check if these samples satisfy our constraints
            k_indx = self.summands[1].satisfies_constraints(Xs)
            # Identify the number of new samples to keep from this iteration
            n_it = np.sum(k_indx)
            # Keep iterating if there are not valid samples
            if n_it != 0:
                # Keep valid samples
                Xs = Xs[k_indx]
                # Add them to Xkeep
                Xk[n_keep:n_keep+n_it,:] = Xs
                # Update n_keep
                n_keep += n_it

        # Rescale if applicable
        if not scale:
            Xk *= self.scale

        # Return samples
        return Xk

    #### Basic Fit Method ####

    def fit_simple(self, Y, w=None, assign=True):
        '''Find the mean and covariance of random samples

        Parameters
        ----------
        Y: array like, shape = (npts, ndim)
            Input points that describe gaussian
        w: array_like, shape = (npts,), optional
            Input weights for points that describe gaussian
        assign: bool, optional
            Input Assign guess to MultivariateNormal Object?
        '''
        ## Imports ##
        import numpy as np

        # Find the average and covariance of the samples
        norm=self.read_guess()[0]
        Y = Y/self.scale
        mean = np.average(Y, weights = w, axis = 0)
        cov = np.cov(Y.T, aweights = w)

        # convert to a sample
        X = params_of_norm_mu_cov(norm, mean, cov)

        # Assign this guess
        if assign:
            self.assign_guess(X)

        # Return the likelihood
        return X

    ######## Save ########
    def save(self, fname, label, attrs=None):
        ''' Save the parameters of the fit

        Parameters
        ----------
        fname: str
            Input path to file location
        label: str
            Input fit label within file
        attrs: dict, optional
            additional attributes to save
        '''
        # Open Database pointer
        db = Database(fname, group=label)
        # Calculate things
        norm = self.read_norm()
        mu, cov = self.read_physical()
        cor = self.read_cor()
        std = self.read_std()*self.scale
        limits = self.read_physical_limits()
        #summands[1].read_limits()*self.scale[:,None]
        # Assign things
        db.dset_set("norm",norm)
        db.dset_set("mean",mu)
        db.dset_set("std",std)
        db.dset_set("cor",cor)
        db.dset_set("cov",cov)
        db.dset_set("scale",self.scale)
        db.dset_set("limits",limits)
        # Assign attrs
        db.attr_set(".","seed",self.seed)
        db.attr_set(".","sig_max",self.sig_max)
        db.attr_set("norm","max",self._parameters["norm"].limits[1])
        if not (attrs is None):
            for item in attrs:
                db.attr_set(".",item,attrs[item])
        # Get names
        names = []
        labels = []
        for i, key in enumerate(self.summands[1]._parameters):
            names.append(self.summands[1]._parameters[key].name)
            labels.append(self.summands[1]._parameters[key].label)
        db.attr_set(".", "names", names)
        db.attr_set(".", "labels", labels)

    def exists(self, fname, label):
        ''' Check if fit exists
        Parameters
        ----------
        fname: str
            Input path to file location
        label: str
            Input fit label within file
        '''
        # Imports 
        from os.path import isfile
        # Check if database exists
        if not isfile(fname):
            return False
        # Open Database pointer
        db = Database(fname)
        return db.exists(label)
        


    ######## Load ########
    @staticmethod
    def load(fname, label, normalize=False):
        '''Load the parameters for a particular fit

        Parameters
        ----------
        fname: str
            Input path to file location
        label: str
            Input fit label within file
        '''
        from os.path import isfile
        # Check if release file exists
        if not isfile(fname):
            raise RuntimeError("No such file: %s"%fname)
        db = Database(fname)
        assert db.exists(label)
        db = Database(fname, group=label)
        
        # load norm param
        try:
            norm = db.dset_value("norm")
            # Load norm max
            norm_max = db.attr_value("norm","max")
        except:
            warnings.warn(
                "Warning: no normalization constant found. Assuming 0.0")
            norm = 0.0
            norm_max=DEFAULT_NORM_MAX
        # Load mu params
        mu = db.dset_value("mean")
        # load sig params
        sig = db.dset_value("std")
        # load correlation params
        cor = db.dset_value("cor")
        # load covariance
        cov = db.dset_value("cov")
        # load limits
        limits = db.dset_value("limits")
        # load scale
        scale = db.dset_value("scale")
        # load attrs
        attr_dict = db.attr_dict(".")

        # Assign dimensionality
        ndim = mu.size

        # Initialize list of variable parameters
        variables = []
        for i in range(ndim):
            # Generate parameter names
            if "names" in attr_dict:
                name = attr_dict["names"][i]
            else:
                name = "p%d"%(i)
            # Generate parameter labels
            if "labels" in attr_dict:
                label = attr_dict["labels"][i]
            else:
                label = None
            # Generate parameter
            variables.append(Parameter(name, mu[i], limits[i], label))

        # Check random state
        if "seed" in attr_dict:
            seed = attr_dict["seed"]
        else:
            seed = None

        # Check sig_max
        if "sig_max" in attr_dict:
            sig_max = attr_dict["sig_max"]
        else:
            sig_max = DEFAULT_SIG_MAX

        # Create norm parameter
        pnorm = Parameter("norm", norm, [0.,norm_max],"norm")

        # Initialize object
        MV = MultivariateNormal(
                                variables,
                                scale,
                                seed=seed,
                                sig_max=sig_max,
                                norm=pnorm,
                               )

        # Set params
        _mu_scaled = mu/scale
        _Sig_scaled = cov/np.outer(scale,scale)
        params = params_of_norm_mu_cov(norm, _mu_scaled, _Sig_scaled).flatten()

        # Assign params
        MV.assign_guess(params)

        # Normalize
        if normalize:
            MV.normalize()

        return MV



''' Coordinate tranformations and labelling'''

######## Coordinate Labels ########
coord_labels = {
                "log_likelihood"       : "$lnL$",
                "cos_tilt_1"           : r"$\cos\theta_1$",
                "cos_tilt_2"           : r"$\cos\theta_2$",
                "redshift"             : "$z$",
                "chirp_mass_source"    : "$\mathcal{M}_{c,s}$",
                "symmetric_mass_ratio" : "$\eta$",
                "mass_1"               : "$m_1$",
                "mass_2"               : "$m_2$",
                "ra"                   : "ra",
                "dec"                  : "dec",
                "mass_1_source"        : "$m_{1,s}$",
                "mass_2_source"        : "$m_{2,s}$",
                "phi_1"                : "$\phi_1$",
                "phi_2"                : "$\phi_2$",
                "spin_1x"              : "$\chi_{1x}$",
                "spin_2x"              : "$\chi_{2x}$",
                "spin_1y"              : "$\chi_{1y}$",
                "spin_2y"              : "$\chi_{2y}$",
                "spin_1z"              : "$\chi_{1z}$",
                "spin_2z"              : "$\chi_{2z}$",
                "spin_1xy"             : "$\chi_{1xy}$",
                "spin_2xy"             : "$\chi_{2xy}$",
                "phi_1"                : "$\phi_1$",
                "phi_2"                : "$\phi_2$",
                "phi_12"               : "$\phi_{12}$",
                "a_1"                  : "$a_1$",
                "a_2"                  : "$a_2$",
                "luminosity_distance"  : "$\ell$",
                "inv_lum_dist"         : "$1/\ell$",
                "chirp_mass"           : "$\mathcal{M}_{c,z}$",
                "chi_eff"              : "$\chi_{\mathrm{eff}}$",
                "chi_Minus"            : "$\chi_{m}$",
                "total_mass_source"    : "$M_s$",
                "total_mass"           : "$M$",
                "cos_theta_jn"         : r"$\cos\theta_{jn}$",
                "lambda_1"             : "$\Lambda_1$",
                "lambda_2"             : "$\Lambda_2$",
                "lambda_tilde"         : r"$\tilde{\Lambda}$",
                "delta_lambda"         : r"$\Delta \tilde{\Lambda}$",
                "mass_ratio"           : "$q$",
               } 

coord_tags = {
              "aligned3d"               : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "chi_eff",
                                         "prior_aligned3d",
                                          ),
              "aligned3d_source"        : (
                                         "chirp_mass_source",
                                         "symmetric_mass_ratio",
                                         "chi_eff",
                                         "prior_aligned3d",
                                          ),
              "aligned3d_dist"          : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "chi_eff",
                                         "inv_lum_dist",
                                         "prior_aligned3d_dist",
                                          ),
              "mass_tides"              : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "prior_mass",
                                          ),
              "mass_tides_source"       : (
                                         "chirp_mass_source",
                                         "symmetric_mass_ratio",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "prior_mass",
                                          ),
              "aligned_tides"           : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "chi_eff",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "prior_aligned3d",
                                          ),
              "aligned_tides_source"    : (
                                         "chirp_mass_source",
                                         "symmetric_mass_ratio",
                                         "chi_eff",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "prior_aligned3d",
                                          ),
              "aligned_tides_dist"      : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "chi_eff",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "inv_lum_dist",
                                         "prior_aligned3d_dist",
                                          ),
              "spin6d"                  : (
                                         "spin_1x",
                                         "spin_2x",
                                         "spin_1y",
                                         "spin_2y",
                                         "spin_1z",
                                         "spin_2z",
                                         "prior_precessing8d",
                                          ),
              "precessing8d"            : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "spin_1x",
                                         "spin_2x",
                                         "spin_1y",
                                         "spin_2y",
                                         "spin_1z",
                                         "spin_2z",
                                         "prior_precessing8d",
                                          ),
              "precessing8d_source"     : (
                                         "chirp_mass_source",
                                         "symmetric_mass_ratio",
                                         "spin_1x",
                                         "spin_2x",
                                         "spin_1y",
                                         "spin_2y",
                                         "spin_1z",
                                         "spin_2z",
                                         "prior_precessing8d",
                                          ),
              "precessing8d_dist"       : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "spin_1x",
                                         "spin_2x",
                                         "spin_1y",
                                         "spin_2y",
                                         "spin_1z",
                                         "spin_2z",
                                         "inv_lum_dist",
                                         "prior_precessing8d_dist",
                                          ),
              "precessing_tides_source" : (
                                         "chirp_mass_source",
                                         "symmetric_mass_ratio",
                                         "spin_1x",
                                         "spin_2x",
                                         "spin_1y",
                                         "spin_2y",
                                         "spin_1z",
                                         "spin_2z",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "prior_precessing8d_dist",
                                          ),
              "full_precessing_tides"   : (
                                         "chirp_mass",
                                         "symmetric_mass_ratio",
                                         "spin_1x",
                                         "spin_2x",
                                         "spin_1y",
                                         "spin_2y",
                                         "spin_1z",
                                         "spin_2z",
                                         "lambda_tilde",
                                         "delta_lambda",
                                         "inv_lum_dist",
                                         "prior_precessing8d_dist",
                                          ),
             }




######## Cooridnate Transforms ########

def mc_eta_of_m1_m2(m1, m2):
    '''Get mc and eta from m1 and m2
    Parameters
    ----------
    m1: array like, shape = (npts,)
        Input mass_1 values
    m2: array like, shape = (npts,)
        Input mass_2 values
    '''
    #M = m1 + m2
    mc = (m1*m2)**(3./5.)*(m1+m2)**(-1./5.)
    eta = m1*m2/(m1+m2)/(m1+m2)
    return mc, eta

def m1_m2_of_mc_eta(mc, eta):
    '''Get m1 and m2 back from mc and eta
    Parameters
    ----------
    mc: array like, shape = (npts,)
        Input chirp_mass values
    eta: array like, shape = (npts,)
        Input eta values
    '''
    M = mc*(eta**-0.6)
    m1 = (M/2.)*(1. + (1. - 4.*eta)**(1./2.))
    m2 = (M/2.)*(1. - (1. - 4.*eta)**(1./2.))
    return m1, m2

def q_of_mc_eta(mc, eta):
    '''Get q back from mc and eta
    Parameters
    ----------
    mc: array like, shape = (npts,)
        Input chirp_mass values
    eta: array like, shape = (npts,)
        Input eta values
    '''
    m1, m2 = m1_m2_of_mc_eta(mc, eta)
    return m2/m1

def z_of_lum_dist(lum_dist):
    '''Calculate redshift from luminosity distance (Mpc) using astropy
    Parameters
    ----------
    lum_dist: array like, shape = (npts,)
        Input luminosity distance values
    '''
    import numpy as np
    from astropy.cosmology import z_at_value
    from astropy import units as u
    from astropy.cosmology import Planck13

    # Make it an array
    lum_dist = np.asarray(lum_dist)
    # Initialize output array
    result = np.empty_like(lum_dist)
    # Loop
    for i in range(len(lum_dist)):
        result[i] = z_at_value(Planck13.luminosity_distance, lum_dist[i]*u.Mpc)

    return result

def detector_of_source(M_source, z):
    '''Convert detector frame mass to source frame
    Parameters
    ----------
    M_source: array_like, shape = (npts,)
        Input some source frame mass values
    z: array_like, shape = (npts,)
        Input some redshift values
    '''
    return M_source*(z + 1.)

def source_of_detector(M_detector, z):
    '''Convert source frame mass to detector
    Parameters
    ----------
    M_detector: array_like, shape = (npts,)
        Input some detector frame mass values
    z: array_like, shape = (npts,)
        Input some redshift values
    '''
    return M_detector/(z + 1.)

def chieff_of_m1m2s1s2(m1, m2, chi1z, chi2z):
    '''Convert from spin components to chieff
    Parameters
    ----------
    m1: array like, shape = (npts,)
        Input mass_1 values
    m2: array like, shape = (npts,)
        Input mass_2 values
    chi1z: array like, shape = (npts,)
        Input spin values
    chi2z: array like, shape = (npts,)
        Input spin values
    '''
    return ((m1*chi1z) + (m2*chi2z))/(m1 + m2)

def chiMinus_of_m1m2s1s2(m1, m2, chi1z, chi2z):
    '''Convert from spin components to chi Minus
    Parameters
    ----------
    m1: array like, shape = (npts,)
        Input mass_1 values
    m2: array like, shape = (npts,)
        Input mass_2 values
    chi1z: array like, shape = (npts,)
        Input spin values
    chi2z: array like, shape = (npts,)
        Input spin values
    '''
    return ((m1*chi1z) - (m2*chi2z))/(m1 + m2)

def chi1z_chi2z_of_chieff_chiMinus(m1, m2, chieff, chiMinus):
    '''get chi1 and chi2 from chieff and chiminus
    Parameters
    ----------
    m1: array like, shape = (npts,)
        Input mass_1 values
    m2: array like, shape = (npts,)
        Input mass_2 values
    chieff: array like, shape = (npts,)
        Input chi effective values
    chiMinus: array like, shape = (npts,)
        Input chi Minus values
    '''
    import numpy as np
    chi1z = np.power(2*m1,-1.) * (m1 + m2)*(chieff + chiMinus)
    chi2z = np.power(2*m2,-1.) * (m1 + m2)*(chieff - chiMinus)
    return chi1z, chi2z

def chieff_chiMinus_of_chi1z_chi2z(m1, m2, chi1z, chi2z):
    ''' Get chieff and chiminus from chi1z, chi2z
    Parameters
    ----------
    m1: array like, shape = (npts,)
        Input mass_1 values
    m2: array like, shape = (npts,)
        Input mass_2 values
    chi1z: array like, shape = (npts,)
        Input spin values
    chi2z: array like, shape = (npts,)
        Input spin values
    '''
    import numpy as np
    inv_M = np.power(m1 + m2, -1.)
    chieff = inv_M * (chi1z*m1 + chi2z*m2)
    chiMinus = inv_M * (chi1z*m1 - chi2z*m2)
    return chieff, chiMinus

def lam1_lam2_of_pe_params(eta, lambda_tilde, delta_lambda):
    """ get lambda_1 and lambda_2 from lambda_tilde and delta_lambda_tilde
    Parameters
    ----------
    eta: array like, shape = (npts,)
        Input eta values
    lambda_tilde: array_like, shape = (npts,)
        Input lambda tilde values
    delta_lambda: array_like, shape = (npts,)
        Input delta lambda tilde values
    """
    import numpy as np
    a = (8.0/13.0)*(1.0+7.0*eta-31.0*eta**2)
    b = (8.0/13.0)*np.sqrt(1.0-4.0*eta)*(1.0+9.0*eta-11.0*eta**2)
    c = (1.0/2.0)*np.sqrt(1.0-4.0*eta)*(1.0 - 13272.0*eta/1319.0 + 8944.0*eta**2/1319.0)
    d = (1.0/2.0)*(1.0 - 15910.0*eta/1319.0 + 32850.0*eta**2/1319.0 + 3380.0*eta**3/1319.0)
    den = (a+b)*(c-d) - (a-b)*(c+d)
    lambda_1 = ( (c-d)*lambda_tilde - (a-b)*delta_lambda )/den
    lambda_2 = (-(c+d)*lambda_tilde + (a+b)*delta_lambda )/den
    # Adjust lambda_1 and lambda_2 if lambda_1 becomes negative
    # lambda_2 should be adjusted such that lambda_tilde is held fixed
    #    if lambda_1<0:
    #        lambda_1 = 0
    #        lambda_2 = lambda_tilde / (a-b)
    return lambda_1, lambda_2

def deltalambda_of_eta_lam1_lam2(eta, lambda_1, lambda_2):
    """ Get delta lambda tilde from eta, lambda1, lambda2 

    This is the definition found in Les Wade's paper.
    Les has factored out the quantity \sqrt(1-4\eta). It is different from Marc Favata's paper.
    $\delta\tilde\Lambda(\eta, \Lambda_1, \Lambda_2)$.
    Lambda_1 is assumed to correspond to the more massive (primary) star m_1.
    Lambda_2 is for the secondary star m_2.

    Parameters
    ----------
    eta: array like, shape = (npts,)
        Input eta values
    lambda_1: array_like, shape = (npts,)
        Input lambda_1 neturon star deformability
    lambda_2: array_like, shape = (npts,)
        Input lambda_2 neturon star deformability
    """
    import numpy as np
    return (1.0/2.0)*(
        np.sqrt(1.0-4.0*eta)*(1.0 - 13272.0*eta/1319.0 + 8944.0*eta**2/1319.0)*(lambda_1+lambda_2)
        + (1.0 - 15910.0*eta/1319.0 + 32850.0*eta**2/1319.0 + 3380.0*eta**3/1319.0)*(lambda_1-lambda_2)
    )
    
def lambdatilde_of_eta_lam1_lam2(eta, lambda_1, lambda_2):
    """ Get lambda tilde from eta, lambda1, lambda2 
    $\tilde\Lambda(\eta, \Lambda_1, \Lambda_2)$.
    Lambda_1 is assumed to correspond to the more massive (primary) star m_1.
    Lambda_2 is for the secondary star m_2.

    Parameters
    ----------
    eta: array like, shape = (npts,)
        Input eta values
    lambda_1: array_like, shape = (npts,)
        Input lambda_1 neturon star deformability
    lambda_2: array_like, shape = (npts,)
        Input lambda_2 neturon star deformability
    """
    import numpy as np
    return (8.0/13.0)*((1.0+7.0*eta-31.0*eta**2)*(lambda_1+lambda_2) + np.sqrt(1.0-4.0*eta)*(1.0+9.0*eta-11.0*eta**2)*(lambda_1-lambda_2))


######## More complicated stuff ########

def z_of_lum_dist_interp(lum_dist, nbins = 100):
    '''Interpolate z_of_lum_dist for a large data set

    Parameters
    ----------
    lum_dist: array like, shape = (npts,)
        Input luminosity distance values
    nbins: int, optional
        Input number of bins for interpolation
    '''
    import numpy as np
    from gp_api.utils import fit_compact_nd

    # Make lum_dist a numpy array
    lum_dist = np.asarray(lum_dist)

    # Generate the training space
    X_train = np.linspace(np.min(lum_dist), np.max(lum_dist), nbins)
    # Generate the training data
    Y_train = z_of_lum_dist(X_train)

    # Convert the training space
    X_train = X_train.reshape((X_train.size, 1))
    #Y_train = Y_train.reshape((Y_train.size, 1))

    # Generate fast gaussian process model
    gp = fit_compact_nd(X_train, Y_train)
    
    # Evaluate
    lum_dist = lum_dist.reshape((lum_dist.size,1))
    values = gp.mean(lum_dist).flatten()

    return values




#!/usr/bin/env python3
'''\
load_best_fits.py

Vera Del Favero

Load the best fits for each event, for specified coordinate tags
'''
######## Imports ########

#### Public Imports ####
import numpy as np
from scipy.stats import multivariate_normal
import h5py
from os.path import join, isfile, isdir

#### Local Imports ####
from gwalk.data import Database
from gwalk.coordinates import coord_tags
from gwalk import NAL
from gwalk.multivariate_normal_pdf import multivariate_normal_pdf as parallel_likelihood

######## GLOBALS ########

data_location = "/home/xevra/Repos/nal-methods-paper/data/raw/"
releases = ["GWTC-1", "GWTC-2", "GWTC-2p1", "GWTC-3"]
TEXT_OUTPUT = "/home/xevra/Repos/nal-methods-paper/data/raw"

######## Algorithms ########

def get_bestfits():
    #### Initialize Dictionaries ####
    bestgroups = {}
    bestfits = {}
    bestkl = {}
    for coords in coord_tags:
        bestfits[coords] = {}
        bestgroups[coords] = {}
        bestkl[coords] = {}

    #### Check Data ####

    for release in releases:
        # Check that the file exists
        fname_release = join(data_location, release + ".nal.hdf5")
        if not isfile(fname_release):
            raise RuntimeError("No such file as %s"%fname_release)
        # Load the database
        db_release = Database(fname_release)
        # Get a list of events
        events = db_release.list_items()
        # Loop through each event
        for event in events:
            # Identify fit labels
            labels = db_release.list_items(event)
            if len(labels) == 0:
                raise RuntimeError("No fits at all for %s"%event)
        
            #### Initialize aligned3d_source flag ####
            for coords in coord_tags:
                found = False
                best_fit = None
                best_fit_kl = None
                # Check each label
                for label in labels:
                    # Fit location in file
                    fit_addr = join(event,label)
                    # Fields
                    coord_tag, group, fit_method = label.split(':')
                    # Check coord_tag
                    if coord_tag == coords:
                        # Check kl divergence
                        try:
                            fit_kl = float(db_release.attr_value(fit_addr, "kl"))
                            found = True
                            # Check best fits while at it
                            if best_fit is None:
                                best_fit = label
                                best_fit_kl = float(fit_kl)
                            elif fit_kl < best_fit_kl:
                                best_fit = label
                                best_fit_kl = fit_kl
                        except:
                            pass
                if found:
                    # Identify best fit information from label
                    [coord_tag, group, fit_method] = best_fit.split(":")
                    Norm = NAL.load_fit(
                                        fname_release,
                                        group,
                                        event,
                                        coord_tag,
                                        fit_method,
                                       )
                    bestfits[coords][event] = Norm
                    bestgroups[coords][event] = group
                    bestkl[coords][event] = best_fit_kl
                
    return bestfits, bestgroups, bestkl

def write_tag_summary(bestfits, bestgroups, bestkl, coord_tag, outfile):
    '''\
    Write data output to file
    '''
    fit_dict = bestfits[coord_tag]
    fit_groups = bestgroups[coord_tag]
    fit_kl = bestkl[coord_tag]
    coords = coord_tags[coord_tag][:-1]
    ndim = len(coords)
    events = fit_dict.keys()
    with open(outfile, 'w') as F:
        for event in events:
            # Load physical quantities
            mean, cov = fit_dict[event].read_physical()
            # Load these other quantities
            std = np.sqrt(np.diag(cov))
            cor = fit_dict[event].guess_cor()
            # Initialize fields for line
            fields = []
            # Append event name
            fields.append(event)
            # Append group
            fields.append(fit_groups[event])
            # Append kl divergence 
            fields.append(fit_kl[event])
            # Append coordinate names
            for item in coords:
                fields.append(item)
            # Append mean
            for i in range(ndim):
                fields.append("%2.6f"%mean[i])
            # Append standard deviations
            for i in range(ndim):
                fields.append("%2.6f"%std[i])
            # Append correlation factors
            for i in range(ndim):
                for j in range(i):
                    fields.append("%2.6f"%cor[i,j])
            # Generate the line
            line = ""
            for item in fields:
                line = line + "%s, "%item
            line.rstrip(", ")
            line = line + "\n"
            F.writelines(line)

def write_all_bestfits(outdir):
    '''\
    Write all the best fits in text format
    '''
    bestfits, bestgroups, bestkl = get_bestfits()
    for coord_tag in bestfits.keys():
        outfile = join(outdir, "best_nal_fits_%s.txt"%(coord_tag))
        write_tag_summary(bestfits, bestgroups, bestkl, coord_tag, outfile)





######## Main ########

def main():
    write_all_bestfits(TEXT_OUTPUT)

######## Execution ########

if __name__ == "__main__":
    main()
